import unittest
import torch
import os
from bagua.torch_api.communication import (
    _get_default_group,
    allreduce,
    allreduce_inplace,
    send,
    recv,
    allgather,
    barrier,
    broadcast_object,
    reduce,
    reduce_inplace,
    reduce_scatter,
    reduce_scatter_inplace,
    ReduceOp,
)
from tests.internal.common_utils import find_free_port
import multiprocessing
import bagua.torch_api as bagua
import threading
import time
from tests import skip_if_cuda_not_available


class Result(object):
    def __init__(self):
        self.ret = torch.Tensor([False]).bool()
        self.data = torch.Tensor([0.0])


def init_env(rank, env):
    os.environ["WORLD_SIZE"] = env["WORLD_SIZE"]
    os.environ["LOCAL_WORLD_SIZE"] = env["LOCAL_WORLD_SIZE"]
    os.environ["MASTER_ADDR"] = env["MASTER_ADDR"]
    os.environ["MASTER_PORT"] = env["MASTER_PORT"]
    os.environ["BAGUA_SERVICE_PORT"] = env["BAGUA_SERVICE_PORT"]

    os.environ["RANK"] = str(rank)
    os.environ["LOCAL_RANK"] = str(rank)

    # init bagua distributed process group
    torch.cuda.set_device(rank)
    bagua.init_process_group()


def run_abort(rank, nprocs, results, env):
    init_env(rank, env)

    os.environ["NCCL_PROTO"] = "^LL128"

    comm_stream = torch.cuda.Stream()
    comm = _get_default_group().get_global_communicator()

    def abort():
        time.sleep(10)
        comm.abort()

    threading.Thread(target=abort).start()

    with torch.cuda.stream(comm_stream):
        data = torch.rand(10).cuda()

    for _ in range(rank + 1):
        comm.allreduce_inplace(
            data.ensure_bagua_tensor().bagua_backend_tensor(), ReduceOp.AVG
        )

    comm_stream.synchronize()


def run_allreduce(rank, nprocs, results, env):
    init_env(rank, env)

    send_tensor = torch.rand(100).cuda()
    recv_tensor = torch.zeros_like(send_tensor)

    tensor = send_tensor.clone()

    allreduce(send_tensor, recv_tensor)

    torch.distributed.all_reduce(tensor)

    results[rank].ret[0] = torch.equal(recv_tensor, tensor)


def run_p2p(rank, nprocs, results, env):
    init_env(rank, env)

    send_tensor = torch.rand(100).cuda()
    recv_tensor = torch.zeros_like(send_tensor)

    if rank % 2 == 0:
        send(send_tensor, dst=(rank + 1) % nprocs)
        results[rank].data.copy_(torch.norm(send_tensor))
    else:
        recv(recv_tensor, src=(rank - 1 + nprocs) % nprocs)
        results[rank].data.copy_(torch.norm(recv_tensor))


def run_allgather(rank, nprocs, results, env):
    init_env(rank, env)

    send_tensor = torch.rand(100).cuda()
    recv_tensor = torch.zeros(
        [nprocs, 100], device=send_tensor.device, dtype=send_tensor.dtype
    )

    tensor = send_tensor.clone()
    tensor_list = [torch.zeros_like(tensor) for _ in range(nprocs)]

    allgather(send_tensor, recv_tensor)

    torch.distributed.all_gather(tensor_list, tensor)

    ret = True
    for i in range(nprocs):
        ret = ret and torch.equal(recv_tensor[i], tensor_list[i])

    results[rank].ret[0] = ret


def run_barrier(rank, nprocs, results, env):
    init_env(rank, env)
    barrier()


def run_bcastobject(rank, nprocs, results, env):
    init_env(rank, env)

    if rank == 0:
        state_dict = {"lr": 0.02, "weight_decay": 1e-4, "momentum": 0.9}
    else:
        state_dict = {}

    state_dict = broadcast_object(state_dict, 0)

    ret = True
    for i in range(nprocs):
        ret = (
            state_dict["lr"] == 0.02
            and state_dict["weight_decay"] == 1e-4
            and state_dict["momentum"] == 0.9
        )

    results[rank].ret[0] = ret


def run_avg(rank, nprocs, results, env):
    init_env(rank, env)

    def reduce_fn(send_tensor, recv_tensor, op):
        reduce(send_tensor, recv_tensor, 0, op)

    def reduce_inplace_fn(tensor, op):
        reduce_inplace(tensor, 0, op)

    fns = [reduce_fn, allreduce]
    inplace_fns = [reduce_inplace_fn, allreduce_inplace]

    succ = True
    for fn in fns:
        send_tensor = torch.rand(100).cuda()
        recv_tensor = torch.zeros_like(send_tensor)

        send_tensor_clone = send_tensor.clone().detach()
        recv_tensor_clone = recv_tensor.clone().detach()

        fn(send_tensor, recv_tensor, op=ReduceOp.AVG)
        fn(send_tensor_clone, recv_tensor_clone, op=ReduceOp.SUM)

        recv_tensor_clone /= nprocs

        torch.cuda.synchronize()
        succ = succ and torch.equal(recv_tensor, recv_tensor_clone)

    for fn in inplace_fns:
        tensor = torch.rand(100).cuda()
        tensor_clone = tensor.clone().detach()

        fn(tensor, op=ReduceOp.AVG)
        fn(tensor_clone, op=ReduceOp.SUM)

        tensor_clone /= nprocs

        torch.cuda.synchronize()
        succ = succ and torch.equal(tensor, tensor_clone)

    results[rank].ret[0] = succ


def run_reduce_scatter(rank, nprocs, results, env):
    init_env(rank, env)

    send_tensor = torch.rand(100 * nprocs).cuda()
    recv_tensor = torch.rand(100).cuda()

    send_tensor_clone = send_tensor.clone().detach()
    recv_tensor_clone = recv_tensor.clone().detach()

    reduce_scatter(send_tensor, recv_tensor, op=ReduceOp.AVG)
    reduce_scatter(send_tensor_clone, recv_tensor_clone, op=ReduceOp.SUM)

    recv_tensor_clone /= nprocs

    tensor = torch.rand(100 * nprocs).cuda()
    tensor_clone = tensor.clone().detach()

    reduce_scatter_inplace(tensor, op=ReduceOp.AVG)
    reduce_scatter_inplace(tensor_clone, op=ReduceOp.AVG)

    tensor_clone /= nprocs

    torch.cuda.synchronize()

    results[rank].ret[0] = torch.equal(recv_tensor, recv_tensor_clone) and torch.equal(
        tensor, tensor_clone
    )


class TestCommunication(unittest.TestCase):
    def run_test_locally(self, fn):
        nprocs = torch.cuda.device_count()
        env = {
            "WORLD_SIZE": str(nprocs),
            "LOCAL_WORLD_SIZE": str(nprocs),
            "MASTER_ADDR": "127.0.0.1",
            "MASTER_PORT": str(find_free_port(8000, 8100)),
            "BAGUA_SERVICE_PORT": str(find_free_port(9000, 9100)),
        }

        mp = multiprocessing.get_context("spawn")
        results = [Result() for _ in range(nprocs)]
        processes = []
        for i in range(nprocs):
            p = mp.Process(
                target=fn,
                args=(i, nprocs, results, env),
            )
            p.start()
            processes.append(p)

        for p in processes:
            p.join(timeout=60)
            self.assertTrue(p.exitcode == 0)

        return results

    @skip_if_cuda_not_available()
    def test_abort(self):
        self.run_test_locally(run_abort)

    @skip_if_cuda_not_available()
    def test_allreduce(self):
        results = self.run_test_locally(run_allreduce)
        for ret in results:
            self.assertTrue(ret.ret.item())

    @skip_if_cuda_not_available()
    def test_p2p(self):
        results = self.run_test_locally(run_p2p)

        i = 1
        while i < len(results):
            self.assertTrue(torch.equal(results[i].data, results[i - 1].data))
            i += 2

    @skip_if_cuda_not_available()
    def test_allgather(self):
        results = self.run_test_locally(run_allgather)
        for ret in results:
            self.assertTrue(ret.ret.item())

    @skip_if_cuda_not_available()
    def test_barrier(self):
        self.run_test_locally(run_barrier)

    @skip_if_cuda_not_available()
    def test_bcastobject(self):
        self.run_test_locally(run_bcastobject)

    @skip_if_cuda_not_available()
    def test_avg(self):
        self.run_test_locally(run_avg)

    @skip_if_cuda_not_available()
    def test_reduce_scatter(self):
        self.run_test_locally(run_reduce_scatter)


if __name__ == "__main__":
    unittest.main()



import torch
import torch.nn as nn
import torch.nn.functional as F
from tests.internal.common_utils import find_free_port
from tests.internal.multi_process import setup_bagua_env
import unittest
import multiprocessing
from bagua.torch_api.utils import flatten
import bagua.torch_api as bagua
from tests import skip_if_cuda_not_available


class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.fc1 = nn.Linear(2, 10, bias=False)
        self.fc2 = nn.Linear(10, 50, bias=True)
        self.fc3 = nn.Linear(50, 4, bias=False)
        self.relu = nn.ReLU()

    def forward(self, x):
        x = self.relu(self.fc1(x))
        x = self.relu(self.fc2(x))
        x = self.fc3(x)
        return F.softmax(x, dim=1)


def _init_bagua_env(rank, env):
    # set deterministic
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True
    torch.manual_seed(rank)
    # initialize subprocess env
    setup_bagua_env(rank, env)


def run_model(
    rank,
    nprocs,
    hierarchical,
    results,
    env,
):
    _init_bagua_env(rank, env)

    # construct model and optimizer, etc.
    model = Net().cuda()
    optimizer = torch.optim.SGD(model.parameters(), lr=0.01)
    loss_fn = nn.MSELoss()

    def run_epochs(num_epochs):
        for epoch in range(num_epochs):
            data = torch.randn(4, 2).cuda()
            target = torch.randn(4, 4).cuda()

            optimizer.zero_grad()
            output = model(data)
            loss = loss_fn(output, target)

            loss.backward()
            optimizer.step()

    run_epochs(1)

    # wrap model
    model = model.with_bagua(
        [optimizer],
        bagua.algorithms.gradient_allreduce.GradientAllReduceAlgorithm(
            hierarchical=hierarchical,
        ),
    )

    run_epochs(10)

    ret = results[rank]

    ret._weight.copy_(flatten([param.data for param in model.parameters()]))


class Result(object):
    def __init__(self):
        model = Net()
        self._weight = flatten(
            [torch.zeros_like(param.data) for param in model.parameters()]
        )


class TestGradientAllReduce(unittest.TestCase):
    def run_test_locally(
        self,
        nprocs,
        hierarchical,
    ):
        env = {
            "WORLD_SIZE": str(nprocs),
            "LOCAL_WORLD_SIZE": str(nprocs),
            "MASTER_ADDR": "127.0.0.1",
            "MASTER_PORT": str(find_free_port(8000, 8100)),
            "BAGUA_SERVICE_PORT": str(find_free_port(9000, 9100)),
        }

        mp = multiprocessing.get_context("spawn")
        results = [Result() for _ in range(nprocs)]
        processes = []
        for i in range(nprocs):
            p = mp.Process(
                target=run_model,
                args=(
                    i,
                    nprocs,
                    hierarchical,
                    results,
                    env,
                ),
            )
            p.start()
            processes.append(p)

        for p in processes:
            p.join(timeout=60)
            self.assertTrue(p.exitcode == 0)

        for rank in range(nprocs):
            peer_rank = (rank + 1) % nprocs
            # all workers have equal weights
            self.assertTrue(
                torch.equal(
                    results[rank]._weight,
                    results[peer_rank]._weight,
                )
            )

    @skip_if_cuda_not_available()
    def test_algorithm(self):
        nprocs = torch.cuda.device_count()
        self.run_test_locally(
            nprocs=nprocs,
            hierarchical=False,
        )


if __name__ == "__main__":
    unittest.main()



# coding=utf-8
# Copyright 2018 The Google AI Language Team Authors and The HuggingFace Inc. team.
# Copyright (c) 2018, NVIDIA CORPORATION.
# Modifications copyright (c) 2021 Kwai AI Platform and DS3 Lab.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Finetuning the library models for question-answering on SQuAD (DistilBERT, Bert, XLM, XLNet)."""


import argparse
import glob
import logging
import os
import random
import timeit

import numpy as np
import torch
from torch.utils.data import DataLoader, RandomSampler, SequentialSampler
from torch.utils.data.distributed import DistributedSampler
from tqdm import tqdm, trange

import transformers
from transformers import (
    MODEL_FOR_QUESTION_ANSWERING_MAPPING,
    WEIGHTS_NAME,
    AdamW,
    AutoConfig,
    AutoModelForQuestionAnswering,
    AutoTokenizer,
    get_linear_schedule_with_warmup,
    squad_convert_examples_to_features,
)
from transformers.data.metrics.squad_metrics import (
    compute_predictions_log_probs,
    compute_predictions_logits,
    squad_evaluate,
)
from transformers.data.processors.squad import (
    SquadResult,
    SquadV1Processor,
    SquadV2Processor,
)
from transformers.trainer_utils import is_main_process
import bagua.torch_api as bagua


try:
    from torch.utils.tensorboard import SummaryWriter
except ImportError:
    from tensorboardX import SummaryWriter


logger = logging.getLogger(__name__)

MODEL_CONFIG_CLASSES = list(MODEL_FOR_QUESTION_ANSWERING_MAPPING.keys())
MODEL_TYPES = tuple(conf.model_type for conf in MODEL_CONFIG_CLASSES)


def set_seed(args):
    random.seed(args.seed)
    np.random.seed(args.seed)
    torch.manual_seed(args.seed)
    if args.n_gpu > 0:
        torch.cuda.manual_seed_all(args.seed)


def to_list(tensor):
    return tensor.detach().cpu().tolist()


def train(args, train_dataset, model, tokenizer):
    """Train the model"""
    if bagua.get_local_rank() == 0:
        tb_writer = SummaryWriter()

    args.train_batch_size = args.per_gpu_train_batch_size * max(1, args.n_gpu)
    train_sampler = (
        RandomSampler(train_dataset)
        if not args.distributed
        else DistributedSampler(
            train_dataset, num_replicas=bagua.get_world_size(), rank=bagua.get_rank()
        )
    )
    train_dataloader = DataLoader(
        train_dataset, sampler=train_sampler, batch_size=args.train_batch_size
    )

    if args.max_steps > 0:
        t_total = args.max_steps
        args.num_train_epochs = (
            args.max_steps
            // (len(train_dataloader) // args.gradient_accumulation_steps)
            + 1
        )
    else:
        t_total = (
            len(train_dataloader)
            // args.gradient_accumulation_steps
            * args.num_train_epochs
        )

    # Prepare optimizer and schedule (linear warmup and decay)
    no_decay = ["bias", "LayerNorm.weight"]
    optimizer_grouped_parameters = [
        {
            "params": [
                p
                for n, p in model.named_parameters()
                if not any(nd in n for nd in no_decay)
            ],
            "weight_decay": args.weight_decay,
        },
        {
            "params": [
                p
                for n, p in model.named_parameters()
                if any(nd in n for nd in no_decay)
            ],
            "weight_decay": 0.0,
        },
    ]
    optimizer = AdamW(
        optimizer_grouped_parameters, lr=args.learning_rate, eps=args.adam_epsilon
    )

    if args.fuse_optimizer:
        optimizer = bagua.contrib.fuse_optimizer(optimizer)

    if args.algorithm == "gradient_allreduce":
        from bagua.torch_api.algorithms import gradient_allreduce

        algorithm = gradient_allreduce.GradientAllReduceAlgorithm()
    elif args.algorithm == "decentralized":
        from bagua.torch_api.algorithms import decentralized

        algorithm = decentralized.DecentralizedAlgorithm()
    elif args.algorithm == "low_precision_decentralized":
        from bagua.torch_api.algorithms import decentralized

        algorithm = decentralized.LowPrecisionDecentralizedAlgorithm()
    elif args.algorithm == "bytegrad":
        from bagua.torch_api.algorithms import bytegrad

        algorithm = bytegrad.ByteGradAlgorithm()
    elif args.algorithm == "qadam":
        from bagua.torch_api.algorithms import q_adam

        optimizer = q_adam.QAdamOptimizer(
            model.parameters(),
            lr=args.learning_rate,
            warmup_steps=100,
            eps=args.adam_epsilon,
        )
        algorithm = q_adam.QAdamAlgorithm(optimizer)
    elif args.algorithm == "async":
        from bagua.torch_api.algorithms import async_model_average

        algorithm = async_model_average.AsyncModelAverageAlgorithm(
            sync_interval_ms=args.async_sync_interval,
            warmup_steps=args.async_warmup_steps,
        )
    else:
        raise NotImplementedError

    scheduler = get_linear_schedule_with_warmup(
        optimizer, num_warmup_steps=args.warmup_steps, num_training_steps=t_total
    )

    # Check if saved optimizer or scheduler states exist
    if os.path.isfile(
        os.path.join(args.model_name_or_path, "optimizer.pt")
    ) and os.path.isfile(os.path.join(args.model_name_or_path, "scheduler.pt")):
        # Load in optimizer and scheduler states
        optimizer.load_state_dict(
            torch.load(os.path.join(args.model_name_or_path, "optimizer.pt"))
        )
        scheduler.load_state_dict(
            torch.load(os.path.join(args.model_name_or_path, "scheduler.pt"))
        )

    if args.fp16:
        try:
            from apex import amp
        except ImportError:
            raise ImportError(
                "Please install apex from https://www.github.com/nvidia/apex to use fp16 training."
            )

        model, optimizer = amp.initialize(
            model, optimizer, opt_level=args.fp16_opt_level
        )

    # multi-gpu training (should be after apex fp16 initialization)
    if args.n_gpu > 1:
        model = torch.nn.DataParallel(model)

    # Distributed training (should be after apex fp16 initialization)
    if args.distributed:
        model = model.with_bagua(
            [optimizer], algorithm, do_flatten=not args.fuse_optimizer
        )

    # Train!
    logger.info("***** Running training *****")
    logger.info("  Num examples = %d", len(train_dataset))
    logger.info("  Num Epochs = %d", args.num_train_epochs)
    logger.info(
        "  Instantaneous batch size per GPU = %d", args.per_gpu_train_batch_size
    )
    logger.info(
        "  Total train batch size (w. parallel, distributed & accumulation) = %d",
        args.train_batch_size
        * args.gradient_accumulation_steps
        * bagua.get_world_size(),
    )
    logger.info("  Gradient Accumulation steps = %d", args.gradient_accumulation_steps)
    logger.info("  Total optimization steps = %d", t_total)

    global_step = 1
    epochs_trained = 0
    steps_trained_in_current_epoch = 0
    # Check if continuing training from a checkpoint
    if os.path.exists(args.model_name_or_path):
        try:
            # set global_step to gobal_step of last saved checkpoint from model path
            checkpoint_suffix = args.model_name_or_path.split("-")[-1].split("/")[0]
            global_step = int(checkpoint_suffix)
            epochs_trained = global_step // (
                len(train_dataloader) // args.gradient_accumulation_steps
            )
            steps_trained_in_current_epoch = global_step % (
                len(train_dataloader) // args.gradient_accumulation_steps
            )

            logger.info(
                "  Continuing training from checkpoint, will skip to saved global_step"
            )
            logger.info("  Continuing training from epoch %d", epochs_trained)
            logger.info("  Continuing training from global step %d", global_step)
            logger.info(
                "  Will skip the first %d steps in the first epoch",
                steps_trained_in_current_epoch,
            )
        except ValueError:
            logger.info("  Starting fine-tuning.")

    tr_loss, logging_loss = 0.0, 0.0
    model.zero_grad()
    train_iterator = trange(
        epochs_trained,
        int(args.num_train_epochs),
        desc="Epoch",
        disable=bagua.get_rank() != 0,
    )

    for _ in train_iterator:
        epoch_iterator = tqdm(
            train_dataloader, desc="Iteration", disable=bagua.get_rank() != 0
        )
        for step, batch in enumerate(epoch_iterator):
            if args.prof >= 0 and step == args.prof:
                print("Profiling begun at iteration {}".format(step))
                torch.cuda.cudart().cudaProfilerStart()

            if args.prof >= 0:
                torch.cuda.nvtx.range_push("Body of iteration {}".format(step))

            # Skip past any already trained steps if resuming training
            if steps_trained_in_current_epoch > 0:
                steps_trained_in_current_epoch -= 1
                continue

            model.train()
            batch = tuple(t.to(args.device) for t in batch)

            inputs = {
                "input_ids": batch[0],
                "attention_mask": batch[1],
                "token_type_ids": batch[2],
                "start_positions": batch[3],
                "end_positions": batch[4],
            }

            if args.model_type in [
                "xlm",
                "roberta",
                "distilbert",
                "camembert",
                "bart",
                "longformer",
            ]:
                del inputs["token_type_ids"]

            if args.model_type in ["xlnet", "xlm"]:
                inputs.update({"cls_index": batch[5], "p_mask": batch[6]})
                if args.version_2_with_negative:
                    inputs.update({"is_impossible": batch[7]})
                if hasattr(model, "config") and hasattr(model.config, "lang2id"):
                    inputs.update(
                        {
                            "langs": (
                                torch.ones(batch[0].shape, dtype=torch.int64)
                                * args.lang_id
                            ).to(args.device)
                        }
                    )

            if args.prof >= 0:
                torch.cuda.nvtx.range_push("forward")

            outputs = model(**inputs)

            if args.prof >= 0:
                torch.cuda.nvtx.range_pop()

            if args.prof >= 0:
                torch.cuda.nvtx.range_push("backward")

            # model outputs are always tuple in transformers (see doc)
            loss = outputs[0]

            if args.n_gpu > 1:
                loss = (
                    loss.mean()
                )  # mean() to average on multi-gpu parallel (not distributed) training
            if args.gradient_accumulation_steps > 1:
                loss = loss / args.gradient_accumulation_steps

            if args.fp16:
                with amp.scale_loss(loss, optimizer) as scaled_loss:
                    scaled_loss.backward()
            else:
                loss.backward()
            if args.prof >= 0:
                torch.cuda.nvtx.range_pop()

            tr_loss += loss.item()

            if (step + 1) % args.gradient_accumulation_steps == 0:
                if args.prof >= 0:
                    torch.cuda.nvtx.range_push("optimizer.step()")

                if args.fp16:
                    torch.nn.utils.clip_grad_norm_(
                        amp.master_params(optimizer), args.max_grad_norm
                    )
                else:
                    torch.nn.utils.clip_grad_norm_(
                        model.parameters(), args.max_grad_norm
                    )

                if args.fuse_optimizer:
                    optimizer.fuse_step()
                else:
                    optimizer.step()

                if args.prof >= 0:
                    torch.cuda.nvtx.range_pop()

                scheduler.step()  # Update learning rate schedule
                model.zero_grad()
                global_step += 1

                # Log metrics
                if (
                    bagua.get_rank() == 0
                    and args.logging_steps > 0
                    and global_step % args.logging_steps == 0
                ):
                    # Only evaluate when single GPU otherwise metrics may not average well
                    if not args.distributed and args.evaluate_during_training:
                        results = evaluate(args, model, tokenizer)
                        for key, value in results.items():
                            tb_writer.add_scalar(
                                "eval_{}".format(key), value, global_step
                            )
                    tb_writer.add_scalar("lr", scheduler.get_lr()[0], global_step)
                    tb_writer.add_scalar(
                        "loss",
                        (tr_loss - logging_loss) / args.logging_steps,
                        global_step,
                    )
                    logging_loss = tr_loss

                # Save model checkpoint
                if (
                    bagua.get_rank() == 0
                    and args.save_steps > 0
                    and global_step % args.save_steps == 0
                ):
                    output_dir = os.path.join(
                        args.output_dir, "checkpoint-{}".format(global_step)
                    )
                    # Take care of distributed/parallel training
                    model_to_save = model.module if hasattr(model, "module") else model
                    model_to_save.save_pretrained(output_dir)
                    tokenizer.save_pretrained(output_dir)

                    torch.save(args, os.path.join(output_dir, "training_args.bin"))
                    logger.info("Saving model checkpoint to %s", output_dir)

                    torch.save(
                        optimizer.state_dict(), os.path.join(output_dir, "optimizer.pt")
                    )
                    torch.save(
                        scheduler.state_dict(), os.path.join(output_dir, "scheduler.pt")
                    )
                    logger.info(
                        "Saving optimizer and scheduler states to %s", output_dir
                    )

            # Pop range "Body of iteration {}".format(i)
            if args.prof >= 0:
                torch.cuda.nvtx.range_pop()

            if args.prof >= 0 and step == args.prof + 10:
                print("Profiling ended at iteration {}".format(step))
                torch.cuda.cudart().cudaProfilerStop()

                if args.algorithm == "async":
                    model.bagua_algorithm.abort(model)
                quit()

            if args.max_steps > 0 and global_step > args.max_steps:
                epoch_iterator.close()
                break
        if args.max_steps > 0 and global_step > args.max_steps:
            train_iterator.close()
            break

    if bagua.get_rank() == 0:
        tb_writer.close()

    if args.algorithm == "async":
        model.bagua_algorithm.abort(model)

    return global_step, tr_loss / global_step


def evaluate(args, model, tokenizer, prefix=""):
    dataset, examples, features = load_and_cache_examples(
        args, tokenizer, evaluate=True, output_examples=True
    )

    if not os.path.exists(args.output_dir) and bagua.get_rank() == 0:
        os.makedirs(args.output_dir)

    args.eval_batch_size = args.per_gpu_eval_batch_size * max(1, args.n_gpu)

    # Note that DistributedSampler samples randomly
    eval_sampler = SequentialSampler(dataset)
    eval_dataloader = DataLoader(
        dataset, sampler=eval_sampler, batch_size=args.eval_batch_size
    )

    # multi-gpu evaluate
    if args.n_gpu > 1 and not isinstance(model, torch.nn.DataParallel):
        model = torch.nn.DataParallel(model)

    # Eval!
    logger.info("***** Running evaluation {} *****".format(prefix))
    logger.info("  Num examples = %d", len(dataset))
    logger.info("  Batch size = %d", args.eval_batch_size)

    all_results = []
    start_time = timeit.default_timer()

    for batch in tqdm(eval_dataloader, desc="Evaluating"):
        model.eval()
        batch = tuple(t.to(args.device) for t in batch)

        with torch.no_grad():
            inputs = {
                "input_ids": batch[0],
                "attention_mask": batch[1],
                "token_type_ids": batch[2],
            }

            if args.model_type in [
                "xlm",
                "roberta",
                "distilbert",
                "camembert",
                "bart",
                "longformer",
            ]:
                del inputs["token_type_ids"]

            feature_indices = batch[3]

            # XLNet and XLM use more arguments for their predictions
            if args.model_type in ["xlnet", "xlm"]:
                inputs.update({"cls_index": batch[4], "p_mask": batch[5]})
                # for lang_id-sensitive xlm models
                if hasattr(model, "config") and hasattr(model.config, "lang2id"):
                    inputs.update(
                        {
                            "langs": (
                                torch.ones(batch[0].shape, dtype=torch.int64)
                                * args.lang_id
                            ).to(args.device)
                        }
                    )
            outputs = model(**inputs)

        for i, feature_index in enumerate(feature_indices):
            eval_feature = features[feature_index.item()]
            unique_id = int(eval_feature.unique_id)

            output = [to_list(output[i]) for output in outputs.to_tuple()]

            # Some models (XLNet, XLM) use 5 arguments for their predictions, while the other "simpler"
            # models only use two.
            if len(output) >= 5:
                start_logits = output[0]
                start_top_index = output[1]
                end_logits = output[2]
                end_top_index = output[3]
                cls_logits = output[4]

                result = SquadResult(
                    unique_id,
                    start_logits,
                    end_logits,
                    start_top_index=start_top_index,
                    end_top_index=end_top_index,
                    cls_logits=cls_logits,
                )

            else:
                start_logits, end_logits = output
                result = SquadResult(unique_id, start_logits, end_logits)

            all_results.append(result)

    evalTime = timeit.default_timer() - start_time
    logger.info(
        "  Evaluation done in total %f secs (%f sec per example)",
        evalTime,
        evalTime / len(dataset),
    )

    # Compute predictions
    output_prediction_file = os.path.join(
        args.output_dir, "predictions_{}.json".format(prefix)
    )
    output_nbest_file = os.path.join(
        args.output_dir, "nbest_predictions_{}.json".format(prefix)
    )

    if args.version_2_with_negative:
        output_null_log_odds_file = os.path.join(
            args.output_dir, "null_odds_{}.json".format(prefix)
        )
    else:
        output_null_log_odds_file = None

    # XLNet and XLM use a more complex post-processing procedure
    if args.model_type in ["xlnet", "xlm"]:
        start_n_top = (
            model.config.start_n_top
            if hasattr(model, "config")
            else model.module.config.start_n_top
        )
        end_n_top = (
            model.config.end_n_top
            if hasattr(model, "config")
            else model.module.config.end_n_top
        )

        predictions = compute_predictions_log_probs(
            examples,
            features,
            all_results,
            args.n_best_size,
            args.max_answer_length,
            output_prediction_file,
            output_nbest_file,
            output_null_log_odds_file,
            start_n_top,
            end_n_top,
            args.version_2_with_negative,
            tokenizer,
            args.verbose_logging,
        )
    else:
        predictions = compute_predictions_logits(
            examples,
            features,
            all_results,
            args.n_best_size,
            args.max_answer_length,
            args.do_lower_case,
            output_prediction_file,
            output_nbest_file,
            output_null_log_odds_file,
            args.verbose_logging,
            args.version_2_with_negative,
            args.null_score_diff_threshold,
            tokenizer,
        )

    # Compute the F1 and exact scores.
    results = squad_evaluate(examples, predictions)
    return results


def load_and_cache_examples(args, tokenizer, evaluate=False, output_examples=False):
    if args.distributed and bagua.get_rank() != 0 and not evaluate:
        # Make sure only the first process in distributed training process the dataset, and the others will use the cache
        torch.distributed.barrier()

    # Load data features from cache or dataset file
    input_dir = args.data_dir if args.data_dir else "."
    cached_features_file = os.path.join(
        input_dir,
        "cached_{}_{}_{}".format(
            "dev" if evaluate else "train",
            list(filter(None, args.model_name_or_path.split("/"))).pop(),
            str(args.max_seq_length),
        ),
    )

    # Init features and dataset from cache if it exists
    if os.path.exists(cached_features_file) and not args.overwrite_cache:
        logger.info("Loading features from cached file %s", cached_features_file)
        features_and_dataset = torch.load(cached_features_file)
        features, dataset, examples = (
            features_and_dataset["features"],
            features_and_dataset["dataset"],
            features_and_dataset["examples"],
        )
    else:
        logger.info("Creating features from dataset file at %s", input_dir)

        if not args.data_dir and (
            (evaluate and not args.predict_file)
            or (not evaluate and not args.train_file)
        ):
            try:
                import tensorflow_datasets as tfds
            except ImportError:
                raise ImportError(
                    "If not data_dir is specified, tensorflow_datasets needs to be installed."
                )

            if args.version_2_with_negative:
                logger.warning(
                    "tensorflow_datasets does not handle version 2 of SQuAD."
                )

            tfds_examples = tfds.load("squad")
            examples = SquadV1Processor().get_examples_from_dataset(
                tfds_examples, evaluate=evaluate
            )
        else:
            processor = (
                SquadV2Processor()
                if args.version_2_with_negative
                else SquadV1Processor()
            )
            if evaluate:
                examples = processor.get_dev_examples(
                    args.data_dir, filename=args.predict_file
                )
            else:
                examples = processor.get_train_examples(
                    args.data_dir, filename=args.train_file
                )

        features, dataset = squad_convert_examples_to_features(
            examples=examples,
            tokenizer=tokenizer,
            max_seq_length=args.max_seq_length,
            doc_stride=args.doc_stride,
            max_query_length=args.max_query_length,
            is_training=not evaluate,
            return_dataset="pt",
            threads=args.threads,
        )

        if bagua.get_rank() == 0:
            logger.info("Saving features into cached file %s", cached_features_file)
            torch.save(
                {"features": features, "dataset": dataset, "examples": examples},
                cached_features_file,
            )

    if args.distributed and bagua.get_rank() == 0 and not evaluate:
        # Make sure only the first process in distributed training process the dataset, and the others will use the cache
        torch.distributed.barrier()

    if output_examples:
        return dataset, examples, features
    return dataset


def main():
    parser = argparse.ArgumentParser()

    # Required parameters
    parser.add_argument(
        "--model_type",
        default=None,
        type=str,
        required=True,
        help="Model type selected in the list: " + ", ".join(MODEL_TYPES),
    )
    parser.add_argument(
        "--model_name_or_path",
        default=None,
        type=str,
        required=True,
        help="Path to pretrained model or model identifier from huggingface.co/models",
    )
    parser.add_argument(
        "--output_dir",
        default=None,
        type=str,
        required=True,
        help="The output directory where the model checkpoints and predictions will be written.",
    )

    # Other parameters
    parser.add_argument(
        "--data_dir",
        default=None,
        type=str,
        help="The input data dir. Should contain the .json files for the task."
        + "If no data dir or train/predict files are specified, will run with tensorflow_datasets.",
    )
    parser.add_argument(
        "--train_file",
        default=None,
        type=str,
        help="The input training file. If a data dir is specified, will look for the file there"
        + "If no data dir or train/predict files are specified, will run with tensorflow_datasets.",
    )
    parser.add_argument(
        "--predict_file",
        default=None,
        type=str,
        help="The input evaluation file. If a data dir is specified, will look for the file there"
        + "If no data dir or train/predict files are specified, will run with tensorflow_datasets.",
    )
    parser.add_argument(
        "--config_name",
        default="",
        type=str,
        help="Pretrained config name or path if not the same as model_name",
    )
    parser.add_argument(
        "--tokenizer_name",
        default="",
        type=str,
        help="Pretrained tokenizer name or path if not the same as model_name",
    )
    parser.add_argument(
        "--cache_dir",
        default="",
        type=str,
        help="Where do you want to store the pre-trained models downloaded from huggingface.co",
    )

    parser.add_argument(
        "--version_2_with_negative",
        action="store_true",
        help="If true, the SQuAD examples contain some that do not have an answer.",
    )
    parser.add_argument(
        "--null_score_diff_threshold",
        type=float,
        default=0.0,
        help="If null_score - best_non_null is greater than the threshold predict null.",
    )

    parser.add_argument(
        "--max_seq_length",
        default=384,
        type=int,
        help="The maximum total input sequence length after WordPiece tokenization. Sequences "
        "longer than this will be truncated, and sequences shorter than this will be padded.",
    )
    parser.add_argument(
        "--doc_stride",
        default=128,
        type=int,
        help="When splitting up a long document into chunks, how much stride to take between chunks.",
    )
    parser.add_argument(
        "--max_query_length",
        default=64,
        type=int,
        help="The maximum number of tokens for the question. Questions longer than this will "
        "be truncated to this length.",
    )
    parser.add_argument(
        "--do_train", action="store_true", help="Whether to run training."
    )
    parser.add_argument(
        "--do_eval", action="store_true", help="Whether to run eval on the dev set."
    )
    parser.add_argument(
        "--evaluate_during_training",
        action="store_true",
        help="Run evaluation during training at each logging step.",
    )
    parser.add_argument(
        "--do_lower_case",
        action="store_true",
        help="Set this flag if you are using an uncased model.",
    )

    parser.add_argument(
        "--per_gpu_train_batch_size",
        default=8,
        type=int,
        help="Batch size per GPU/CPU for training.",
    )
    parser.add_argument(
        "--per_gpu_eval_batch_size",
        default=8,
        type=int,
        help="Batch size per GPU/CPU for evaluation.",
    )
    parser.add_argument(
        "--learning_rate",
        default=5e-5,
        type=float,
        help="The initial learning rate for Adam.",
    )
    parser.add_argument(
        "--gradient_accumulation_steps",
        type=int,
        default=1,
        help="Number of updates steps to accumulate before performing a backward/update pass.",
    )
    parser.add_argument(
        "--weight_decay", default=0.0, type=float, help="Weight decay if we apply some."
    )
    parser.add_argument(
        "--adam_epsilon", default=1e-8, type=float, help="Epsilon for Adam optimizer."
    )
    parser.add_argument(
        "--max_grad_norm", default=1.0, type=float, help="Max gradient norm."
    )
    parser.add_argument(
        "--num_train_epochs",
        default=3.0,
        type=float,
        help="Total number of training epochs to perform.",
    )
    parser.add_argument(
        "--max_steps",
        default=-1,
        type=int,
        help="If > 0: set total number of training steps to perform. Override num_train_epochs.",
    )
    parser.add_argument(
        "--warmup_steps", default=0, type=int, help="Linear warmup over warmup_steps."
    )
    parser.add_argument(
        "--n_best_size",
        default=20,
        type=int,
        help="The total number of n-best predictions to generate in the nbest_predictions.json output file.",
    )
    parser.add_argument(
        "--max_answer_length",
        default=30,
        type=int,
        help="The maximum length of an answer that can be generated. This is needed because the start "
        "and end predictions are not conditioned on one another.",
    )
    parser.add_argument(
        "--verbose_logging",
        action="store_true",
        help="If true, all of the warnings related to data processing will be printed. "
        "A number of warnings are expected for a normal SQuAD evaluation.",
    )
    parser.add_argument(
        "--lang_id",
        default=0,
        type=int,
        help="language id of input for language-specific xlm models (see tokenization_xlm.PRETRAINED_INIT_CONFIGURATION)",
    )

    parser.add_argument(
        "--logging_steps", type=int, default=500, help="Log every X updates steps."
    )
    parser.add_argument(
        "--save_steps",
        type=int,
        default=500,
        help="Save checkpoint every X updates steps.",
    )
    parser.add_argument(
        "--eval_all_checkpoints",
        action="store_true",
        help="Evaluate all checkpoints starting with the same prefix as model_name ending and ending with step number",
    )
    parser.add_argument(
        "--no_cuda", action="store_true", help="Whether not to use CUDA when available"
    )
    parser.add_argument(
        "--overwrite_output_dir",
        action="store_true",
        help="Overwrite the content of the output directory",
    )
    parser.add_argument(
        "--overwrite_cache",
        action="store_true",
        help="Overwrite the cached training and evaluation sets",
    )
    parser.add_argument(
        "--seed", type=int, default=42, help="random seed for initialization"
    )

    parser.add_argument(
        "--fp16",
        action="store_true",
        help="Whether to use 16-bit (mixed) precision (through NVIDIA apex) instead of 32-bit",
    )
    parser.add_argument(
        "--fp16_opt_level",
        type=str,
        default="O1",
        help="For fp16: Apex AMP optimization level selected in ['O0', 'O1', 'O2', and 'O3']."
        "See details at https://nvidia.github.io/apex/amp.html",
    )
    parser.add_argument(
        "--server_ip", type=str, default="", help="Can be used for distant debugging."
    )
    parser.add_argument(
        "--server_port", type=str, default="", help="Can be used for distant debugging."
    )
    parser.add_argument(
        "--threads",
        type=int,
        default=1,
        help="multiple threads for converting example to features",
    )
    parser.add_argument(
        "--set-deterministic",
        action="store_true",
        default=False,
        help="set deterministic or not",
    )
    parser.add_argument(
        "--prof", default=-1, type=int, help="Only run 10 iterations for profiling."
    )
    parser.add_argument(
        "--algorithm",
        type=str,
        default="gradient_allreduce",
        help="distributed algorithm: {gradient_allreduce, bytegrad, decentralized, low_precision_decentralized, qadam, async}",
    )
    parser.add_argument(
        "--async-sync-interval",
        default=500,
        type=int,
        help="Model synchronization interval(ms) for async algorithm",
    )
    parser.add_argument(
        "--async-warmup-steps",
        default=100,
        type=int,
        help="Warmup(allreduce) steps for async algorithm",
    )
    parser.add_argument(
        "--fuse-optimizer",
        action="store_true",
        default=False,
        help="fuse optimizer or not",
    )

    args = parser.parse_args()

    if args.doc_stride >= args.max_seq_length - args.max_query_length:
        logger.warning(
            "WARNING - You've set a doc stride which may be superior to the document length in some "
            "examples. This could result in errors when building features from the examples. Please reduce the doc "
            "stride or increase the maximum length to ensure the features are correctly built."
        )

    if (
        os.path.exists(args.output_dir)
        and os.listdir(args.output_dir)
        and args.do_train
        and not args.overwrite_output_dir
    ):
        raise ValueError(
            "Output directory ({}) already exists and is not empty. Use --overwrite_output_dir to overcome.".format(
                args.output_dir
            )
        )

    # Setup distant debugging if needed
    if args.server_ip and args.server_port:
        # Distant debugging - see https://code.visualstudio.com/docs/python/debugging#_attach-to-a-local-script
        import ptvsd

        print("Waiting for debugger attach")
        ptvsd.enable_attach(
            address=(args.server_ip, args.server_port), redirect_output=True
        )
        ptvsd.wait_for_attach()

    # Setup CUDA, GPU & distributed training
    args.distributed = bagua.get_world_size() > 1
    if not args.distributed or args.no_cuda:
        device = torch.device(
            "cuda" if torch.cuda.is_available() and not args.no_cuda else "cpu"
        )
        args.n_gpu = 0 if args.no_cuda else torch.cuda.device_count()
    else:  # Initializes the distributed backend which will take care of sychronizing nodes/GPUs
        torch.cuda.set_device(bagua.get_local_rank())
        device = torch.device("cuda", bagua.get_local_rank())
        bagua.init_process_group()
        args.n_gpu = 1
    args.device = device

    # Setup logging
    logging.basicConfig(
        format="%(asctime)s - %(levelname)s - %(name)s -   %(message)s",
        datefmt="%m/%d/%Y %H:%M:%S",
        level=logging.WARN,
    )

    if bagua.get_rank() == 0:
        logging.getLogger().setLevel(logging.INFO)

    logger.warning(
        "Process rank: %s, device: %s, n_gpu: %s, distributed training: %s, 16-bits training: %s",
        bagua.get_rank(),
        device,
        args.n_gpu,
        args.distributed,
        args.fp16,
    )
    # Set the verbosity to info of the Transformers logger (on main process only):
    if is_main_process(bagua.get_rank()):
        transformers.utils.logging.set_verbosity_info()
        transformers.utils.logging.enable_default_handler()
        transformers.utils.logging.enable_explicit_format()

    # Added here for reproductibility
    if args.set_deterministic:
        # Set seed
        set_seed(args)
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.deterministic = True

    # Load pretrained model and tokenizer
    if args.distributed and bagua.get_rank() != 0:
        # Make sure only the first process in distributed training will download model & vocab
        torch.distributed.barrier()

    args.model_type = args.model_type.lower()
    config = AutoConfig.from_pretrained(
        args.config_name if args.config_name else args.model_name_or_path,
        cache_dir=args.cache_dir if args.cache_dir else None,
    )
    tokenizer = AutoTokenizer.from_pretrained(
        args.tokenizer_name if args.tokenizer_name else args.model_name_or_path,
        do_lower_case=args.do_lower_case,
        cache_dir=args.cache_dir if args.cache_dir else None,
        use_fast=False,  # SquadDataset is not compatible with Fast tokenizers which have a smarter overflow handeling
    )
    model = AutoModelForQuestionAnswering.from_pretrained(
        args.model_name_or_path,
        from_tf=bool(".ckpt" in args.model_name_or_path),
        config=config,
        cache_dir=args.cache_dir if args.cache_dir else None,
    )

    if args.distributed and bagua.get_rank() == 0:
        # Make sure only the first process in distributed training will download model & vocab
        torch.distributed.barrier()

    model.to(args.device)

    logger.info("Training/evaluation parameters %s", args)

    # Before we do anything with models, we want to ensure that we get fp16 execution of torch.einsum if args.fp16 is set.
    # Otherwise it'll default to "promote" mode, and we'll get fp32 operations. Note that running `--fp16_opt_level="O2"` will
    # remove the need for this code, but it is still valid.
    if args.fp16:
        try:
            import apex

            apex.amp.register_half_function(torch, "einsum")
        except ImportError:
            raise ImportError(
                "Please install apex from https://www.github.com/nvidia/apex to use fp16 training."
            )

    # Training
    if args.do_train:
        train_dataset = load_and_cache_examples(
            args, tokenizer, evaluate=False, output_examples=False
        )
        global_step, tr_loss = train(args, train_dataset, model, tokenizer)
        logger.info(" global_step = %s, average loss = %s", global_step, tr_loss)

    # Save the trained model and the tokenizer
    if args.do_train and bagua.get_rank() == 0:
        logger.info("Saving model checkpoint to %s", args.output_dir)
        # Save a trained model, configuration and tokenizer using `save_pretrained()`.
        # They can then be reloaded using `from_pretrained()`
        # Take care of distributed/parallel training
        model_to_save = model.module if hasattr(model, "module") else model
        model_to_save.save_pretrained(args.output_dir)
        tokenizer.save_pretrained(args.output_dir)

        # Good practice: save your training arguments together with the trained model
        torch.save(args, os.path.join(args.output_dir, "training_args.bin"))

        # Load a trained model and vocabulary that you have fine-tuned
        model = AutoModelForQuestionAnswering.from_pretrained(
            args.output_dir
        )  # , force_download=True)

        # SquadDataset is not compatible with Fast tokenizers which have a smarter overflow handeling
        # So we use use_fast=False here for now until Fast-tokenizer-compatible-examples are out
        tokenizer = AutoTokenizer.from_pretrained(
            args.output_dir, do_lower_case=args.do_lower_case, use_fast=False
        )
        model.to(args.device)

    # Evaluation - we can ask to evaluate all the checkpoints (sub-directories) in a directory
    results = {}
    if args.do_eval and bagua.get_rank() == 0:
        if args.do_train:
            logger.info("Loading checkpoints saved during training for evaluation")
            checkpoints = [args.output_dir]
            if args.eval_all_checkpoints:
                checkpoints = list(
                    os.path.dirname(c)
                    for c in sorted(
                        glob.glob(
                            args.output_dir + "/**/" + WEIGHTS_NAME, recursive=True
                        )
                    )
                )

        else:
            logger.info("Loading checkpoint %s for evaluation", args.model_name_or_path)
            checkpoints = [args.model_name_or_path]

        logger.info("Evaluate the following checkpoints: %s", checkpoints)

        for checkpoint in checkpoints:
            # Reload the model
            global_step = checkpoint.split("-")[-1] if len(checkpoints) > 1 else ""
            model = AutoModelForQuestionAnswering.from_pretrained(
                checkpoint
            )  # , force_download=True)
            model.to(args.device)

            # Evaluate
            result = evaluate(args, model, tokenizer, prefix=global_step)

            result = dict(
                (k + ("_{}".format(global_step) if global_step else ""), v)
                for k, v in result.items()
            )
            results.update(result)

    logger.info("Results: {}".format(results))

    return results


if __name__ == "__main__":
    main()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------
import re

project = "Bagua API Documentation"
copyright = "2021, Kuaishou AI Platform and DS3 Lab"
author = "Kuaishou AI Platform and DS3 Lab"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "autoapi.extension",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
]


napoleon_numpy_docstring = True
autoapi_python_class_content = "both"
autodoc_typehints = "description"
autoapi_type = "python"
autoapi_dirs = ["../bagua"]
autoapi_root = "autoapi"
autoapi_template_dir = "_autoapi_templates"
autoapi_ignore = [
    "*/bagua/autotune/*",
    "*/bagua/distributed/*",
    "*/bagua/script/*",
    "*/bagua/service/*",
    "*/bagua/torch_api/tools/*.py",
    "*/bagua/torch_api/exceptions.py",
    "*/bagua/torch_api/utils.py",
    "*/bagua/torch_api/globals.py",
    "*/bagua/version.py",
    "*/bagua/bagua_define.py",
]
autoapi_options = [
    "members",
    "undoc-members",
    "show-inheritance",
    "imported-members",
]
autoapi_member_order = "groupwise"

master_doc = "autoapi/index"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = False

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "bagua-sphinx-theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    "show_powered_by": False,
    "github_user": "BaguaSys",
    "github_repo": "bagua",
    "github_banner": True,
    "show_related": False,
    "note_bg": "#FFF59C",
}

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
html_use_smartypants = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = True


_ignore_methods = [
    "bagua.torch_api.contrib.*LoadBalancingDistributedSampler.shuffle_chunks",
    "bagua.torch_api.contrib.*LoadBalancingDistributedBatchSampler.generate_batches",
    "bagua.torch_api.contrib.utils.store.ClusterStore.*",
    "bagua.torch_api.algorithms.async_model_average.*.reify",
    "bagua.torch_api.algorithms.async_model_average.*.init*",
    "bagua.torch_api.algorithms.async_model_average.*.need_reset",
    "bagua.torch_api.algorithms.async_model_average.*.tensors_to_buckets",
    "bagua.torch_api.algorithms.bytegrad.*.reify",
    "bagua.torch_api.algorithms.bytegrad.*.init*",
    "bagua.torch_api.algorithms.bytegrad.*.need_reset",
    "bagua.torch_api.algorithms.bytegrad.*.tensors_to_buckets",
    "bagua.torch_api.algorithms.decentralized.*.reify",
    "bagua.torch_api.algorithms.decentralized.*.init*",
    "bagua.torch_api.algorithms.decentralized.*.need_reset",
    "bagua.torch_api.algorithms.decentralized.*.tensors_to_buckets",
    "bagua.torch_api.algorithms.gradient_allreduce.*.reify",
    "bagua.torch_api.algorithms.gradient_allreduce.*.init*",
    "bagua.torch_api.algorithms.gradient_allreduce.*.need_reset",
    "bagua.torch_api.algorithms.gradient_allreduce.*.tensors_to_buckets",
    "bagua.torch_api.algorithms.q_adam.*.reify",
    "bagua.torch_api.algorithms.q_adam.*.init*",
    "bagua.torch_api.algorithms.q_adam.*.need_reset",
    "bagua.torch_api.algorithms.q_adam.*.tensors_to_buckets",
]
_ignore_functions = [
    "bagua.torch_api.env.get_autotune_server_addr",
    "bagua.torch_api.env.get_master_addr",
    "bagua.torch_api.env.get_bagua_service_port",
    "bagua.torch_api.env.is_report_metrics_switch_on",
    "bagua.torch_api.env.get_autotune_level",
    "bagua.torch_api.env.get_autotune_max_samples",
    "bagua.torch_api.env.get_autotune_sampling_confidence_time_s",
    "bagua.torch_api.env.get_autotune_warmup_time_s",
    "bagua.torch_api.env.get_is_output_autotune_log",
    "bagua.torch_api.globals.is_initialized",
    "bagua.torch_api.data_parallel.to_bagua_process_group",
    "bagua.torch_api.data_parallel.functional.torch_reduce_op_to_bagua",
]
_ignore_classes = [
    "bagua.torch_api.algorithms.BaguaModule",
    "bagua.torch_api.algorithms.BaguaBucket",
    "bagua.torch_api.algorithms.BaguaTensor",
    "bagua.torch_api.data_parallel.distributed.*",
    "bagua.torch_api.data_parallel.bagua_distributed.*",
]


def skip_methods(app, what, name, obj, skip, options):
    if what == "method":
        for to_ignore in _ignore_methods:
            p = re.compile(to_ignore)
            if p.match(name):
                skip = True
                return skip

    if what == "function":
        for to_ignore in _ignore_functions:
            p = re.compile(to_ignore)
            if p.match(name):
                skip = True
                return skip

    if what == "class":
        for to_ignore in _ignore_classes:
            p = re.compile(to_ignore)
            if p.match(name):
                skip = True
                return skip

    return skip


def setup(sphinx):
    sphinx.connect("autoapi-skip-member", skip_methods)



import os
import random
import sys
import unittest

import torch
import torch.distributed as c10d

if not c10d.is_available():
    print("c10d not available, skipping tests", file=sys.stderr)
    sys.exit(0)

import torch.nn.functional as F
from torch import nn
from tests.internal.torch.common_distributed import (
    MultiProcessTestCase,
    skip_if_lt_x_gpu,
)
from tests.internal.torch.common_utils import (
    run_tests,
    TEST_WITH_TSAN,
)
from . import test_c10d_common

import bagua.torch_api as bagua
from tests.internal.common_utils import find_free_port
from bagua.torch_api.data_parallel.distributed import DistributedDataParallel_V1_9_0 as DistributedDataParallel


@unittest.skipIf(
    TEST_WITH_TSAN,
    "TSAN is not fork-safe since we're forking in a multi-threaded environment",
)
class DistributedDataParallelTest(test_c10d_common.AbstractDistributedDataParallelTest, MultiProcessTestCase):

    def setUp(self):
        super(DistributedDataParallelTest, self).setUp()
        # NCCL_BLOCKING_WAIT overrides NCCL_ASYNC_ERROR_HANDLING hence tests
        # that use NCCL_BLOCKING_WAIT will test it as expected.
        os.environ["NCCL_ASYNC_ERROR_HANDLING"] = "1"
        os.environ.update(
            {
                "MASTER_ADDR": "127.0.0.1",
                "MASTER_PORT": str(find_free_port(8000, 8100)),
                "BAGUA_SERVICE_PORT": str(find_free_port(9000, 9100)),
            }
        )
        self._spawn_processes()

    def _test_find_unused_parameters_kwarg(self, gradient_as_bucket_view=False):
        """
        Note: this test can be sped up by only running it on a CPU module
        once DistributedDataParallel supports them.
        """
        torch.cuda.set_device(self.rank)
        bagua.init_process_group()
        process_group = c10d.distributed_c10d._get_default_group()

        class FindUnusedParametersModule(nn.Module):
            def __init__(self):
                super(FindUnusedParametersModule, self).__init__()
                self.fc1 = nn.Linear(2, 10, bias=False)
                self.fc2 = nn.Linear(10, 4, bias=False)
                self.fc3 = nn.Linear(4, 4, bias=False)
                self.relu = nn.ReLU()

            def forward(self, x):
                x = self.relu(self.fc1(x))
                x = self.relu(self.fc2(x))
                # Return the fc3 module so that the caller can invoke it
                # outside of the forward function. While this is bad practice,
                # we can use it to trigger a reducer error.
                return (F.softmax(x, dim=1), self.fc3)

        device_id = self.rank
        batch_size = 4
        criterion = nn.CrossEntropyLoss()
        input = torch.rand([batch_size, 2], dtype=torch.float).to(device_id)
        target = torch.LongTensor([random.randrange(4) for _ in range(batch_size)]).to(
            device_id
        )

        ddp_model = None

        def test_find_unused_parameters(
            find_unused_parameters
        ):
            model = DistributedDataParallel(
                FindUnusedParametersModule().float().to(device_id),
                device_ids=[device_id],
                process_group=process_group,
                find_unused_parameters=find_unused_parameters,
            )
            nonlocal ddp_model
            ddp_model = model

            output, fc3 = model(input)
            # output = fc3(output)
            loss = criterion(output, target)
            loss.backward()

        test_find_unused_parameters(find_unused_parameters=True)
        bagua_build_params = [name for name,
                              _ in ddp_model.inner.bagua_build_params()]
        self.assertEqual(set(bagua_build_params),
                         set(['fc1.weight', 'fc2.weight']))

        test_find_unused_parameters(find_unused_parameters=False)
        bagua_build_params = [name for name,
                              _ in ddp_model.inner.bagua_build_params()]
        self.assertEqual(set(bagua_build_params), set(
            ['fc1.weight', 'fc2.weight', 'fc3.weight']))

    @skip_if_lt_x_gpu(2)
    def test_find_unused_parameters_kwarg_debug_detail(self):
        os.environ.update(
            {
                "WORLD_SIZE": str(self.world_size),
                "LOCAL_WORLD_SIZE": str(self.world_size),
                "RANK": str(self.rank),
                "LOCAL_RANK": str(self.rank),
            }
        )

        self._test_find_unused_parameters_kwarg()


if __name__ == "__main__":
    assert (
        not torch.cuda._initialized
    ), "test_distributed must not have initialized CUDA context on main process"

    run_tests()



from bagua.torch_api.contrib.cached_dataset import CachedDataset
from torch.utils.data.dataset import Dataset
import numpy as np
import logging
import unittest
from tests import skip_if_cuda_available

logging.basicConfig(level=logging.DEBUG)


class MyDataset(Dataset):
    def __init__(self, size):
        self.size = size
        self.dataset = [(np.random.rand(5, 2), np.random.rand(1)) for _ in range(size)]

    def __getitem__(self, item):
        return self.dataset[item]

    def __len__(self):
        return self.size


class TestCacheDataset(unittest.TestCase):
    def check_dataset(self, dataset, cache_dataset):
        for _ in range(10):
            for _, _ in enumerate(cache_dataset):
                pass

        for i in range(len(dataset)):
            self.assertTrue((dataset[i][0] == cache_dataset[i][0]).all())
            self.assertTrue((dataset[i][1] == cache_dataset[i][1]).all())

    @skip_if_cuda_available()
    def test_redis(self):
        dataset1 = MyDataset(102)
        dataset2 = MyDataset(102)
        cache_dataset1 = CachedDataset(
            dataset1,
            backend="redis",
            dataset_name="d1",
        )
        cache_dataset2 = CachedDataset(
            dataset2,
            backend="redis",
            dataset_name="d2",
        )

        cache_dataset1.cache_loader.store.clear()

        self.check_dataset(dataset1, cache_dataset1)
        self.assertEqual(cache_dataset1.cache_loader.num_keys(), len(dataset1))

        self.check_dataset(dataset2, cache_dataset2)
        self.assertEqual(
            cache_dataset2.cache_loader.num_keys(), len(dataset1) + len(dataset2)
        )


if __name__ == "__main__":
    unittest.main()



import argparse
import os
import random
import shutil
import time
import warnings
import logging

import torch
import torch.nn as nn
import torch.nn.parallel
import torch.backends.cudnn as cudnn
import torch.optim
import torch.utils.data
import torch.utils.data.distributed
import torchvision.transforms as transforms
import torchvision.datasets as datasets
import torchvision.models as models
import bagua.torch_api as bagua
from bisect import bisect_right


model_names = sorted(
    name
    for name in models.__dict__
    if name.islower() and not name.startswith("__") and callable(models.__dict__[name])
)

parser = argparse.ArgumentParser(description="PyTorch ImageNet Training")
parser.add_argument("data", metavar="DIR", help="path to dataset")
parser.add_argument(
    "-a",
    "--arch",
    metavar="ARCH",
    default="resnet18",
    choices=model_names,
    help="model architecture: " + " | ".join(model_names) + " (default: resnet18)",
)
parser.add_argument(
    "-j",
    "--workers",
    default=4,
    type=int,
    metavar="N",
    help="number of data loading workers (default: 4)",
)
parser.add_argument(
    "--epochs", default=90, type=int, metavar="N", help="number of total epochs to run"
)
parser.add_argument(
    "--warmup-epochs", type=float, default=5, help="number of warmup epochs"
)
parser.add_argument(
    "--start-epoch",
    default=0,
    type=int,
    metavar="N",
    help="manual epoch number (useful on restarts)",
)
parser.add_argument(
    "-b",
    "--batch-size",
    default=256,
    type=int,
    metavar="N",
    help="mini-batch size (default: 256), this is the total "
    "batch size of all GPUs on the current node when "
    "using Data Parallel or Distributed Data Parallel",
)
parser.add_argument(
    "--lr",
    "--learning-rate",
    default=0.1,
    type=float,
    metavar="LR",
    help="initial learning rate",
    dest="lr",
)
parser.add_argument("--momentum", default=0.9, type=float, metavar="M", help="momentum")
parser.add_argument(
    "--wd",
    "--weight-decay",
    default=1e-4,
    type=float,
    metavar="W",
    help="weight decay (default: 1e-4)",
    dest="weight_decay",
)
parser.add_argument(
    "--milestones",
    default="60,70,80",
    type=str,
    help="multi-step learning rate scheduler milestones",
)
parser.add_argument(
    "--gama",
    type=float,
    default=0.2,
    help="multiplicative factor of learning rate decay",
)
parser.add_argument(
    "-p",
    "--print-freq",
    default=10,
    type=int,
    metavar="N",
    help="print frequency (default: 10)",
)
parser.add_argument(
    "--resume",
    default="",
    type=str,
    metavar="PATH",
    help="path to latest checkpoint (default: none)",
)
parser.add_argument(
    "--save-checkpoint", action="store_true", default=False, help="save checkpoint"
)
parser.add_argument(
    "-e",
    "--evaluate",
    dest="evaluate",
    action="store_true",
    help="evaluate model on validation set",
)
parser.add_argument(
    "--pretrained", dest="pretrained", action="store_true", help="use pre-trained model"
)
parser.add_argument(
    "--seed", default=None, type=int, help="seed for initializing training. "
)
parser.add_argument(
    "--amp",
    action="store_true",
    default=False,
    help="use amp",
)

parser.add_argument(
    "--prof", default=-1, type=int, help="Only run 10 iterations for profiling."
)

parser.add_argument(
    "--algorithm",
    type=str,
    default="gradient_allreduce",
    help="distributed algorithm: {gradient_allreduce, bytegrad, decentralized, low_precision_decentralized, qadam, async}",
)

parser.add_argument(
    "--async-sync-interval",
    default=500,
    type=int,
    help="Model synchronization interval(ms) for async algorithm",
)

parser.add_argument(
    "--async-warmup-steps",
    default=100,
    type=int,
    help="Warmup(allreduce) steps for async algorithm",
)

best_acc1 = 0


def main():
    args = parser.parse_args()

    if args.seed is not None:
        random.seed(args.seed)
        torch.manual_seed(args.seed)
        cudnn.deterministic = True
        warnings.warn(
            "You have chosen to seed training. "
            "This will turn on the CUDNN deterministic setting, "
            "which can slow down your training considerably! "
            "You may see unexpected behavior when restarting "
            "from checkpoints."
        )

    torch.cuda.set_device(bagua.get_local_rank())
    bagua.init_process_group()
    args.distributed = bagua.get_world_size() > 1

    logging.basicConfig(
        format="rank-{} %(asctime)s,%(msecs)d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s".format(
            bagua.get_rank()
        ),
        datefmt="%Y-%m-%d:%H:%M:%S",
        level=logging.ERROR,
    )

    if bagua.get_rank() == 0:
        logging.getLogger().setLevel(logging.INFO)

    main_worker(args)


def main_worker(args):
    global best_acc1

    # create model
    if args.pretrained:
        print("=> using pre-trained model '{}'".format(args.arch))
        model = models.__dict__[args.arch](pretrained=True)
    else:
        print("=> creating model '{}'".format(args.arch))
        model = models.__dict__[args.arch]()

    model = model.cuda()

    # define loss function (criterion) and optimizer
    criterion = nn.CrossEntropyLoss().cuda()

    optimizer = torch.optim.SGD(
        model.parameters(),
        args.lr,
        momentum=args.momentum,
        weight_decay=args.weight_decay,
    )

    if args.algorithm == "gradient_allreduce":
        from bagua.torch_api.algorithms import gradient_allreduce

        algorithm = gradient_allreduce.GradientAllReduceAlgorithm()
    elif args.algorithm == "decentralized":
        from bagua.torch_api.algorithms import decentralized

        algorithm = decentralized.DecentralizedAlgorithm()
    elif args.algorithm == "low_precision_decentralized":
        from bagua.torch_api.algorithms import decentralized

        algorithm = decentralized.LowPrecisionDecentralizedAlgorithm()
    elif args.algorithm == "bytegrad":
        from bagua.torch_api.algorithms import bytegrad

        algorithm = bytegrad.ByteGradAlgorithm()
    elif args.algorithm == "qadam":
        from bagua.torch_api.algorithms import q_adam

        optimizer = q_adam.QAdamOptimizer(
            model.parameters(), lr=args.lr, warmup_steps=100
        )
        algorithm = q_adam.QAdamAlgorithm(optimizer)
    elif args.algorithm == "async":
        from bagua.torch_api.algorithms import async_model_average

        algorithm = async_model_average.AsyncModelAverageAlgorithm(
            sync_interval_ms=args.async_sync_interval,
            warmup_steps=args.async_warmup_steps,
        )
    else:
        raise NotImplementedError

    scaler = torch.cuda.amp.GradScaler(enabled=args.amp)

    # optionally resume from a checkpoint
    if args.resume:
        if os.path.isfile(args.resume):
            print("=> loading checkpoint '{}'".format(args.resume))
            # Map model to be loaded to specified single gpu.
            loc = "cuda:{}".format(bagua.get_local_rank())
            checkpoint = torch.load(args.resume, map_location=loc)
            args.start_epoch = checkpoint["epoch"]
            best_acc1 = checkpoint["best_acc1"]
            if bagua.get_local_rank() is not None:
                # best_acc1 may be from a checkpoint from a different GPU
                best_acc1 = best_acc1.to(bagua.get_local_rank())
            model.load_state_dict(checkpoint["state_dict"])
            optimizer.load_state_dict(checkpoint["optimizer"])
            print(
                "=> loaded checkpoint '{}' (epoch {})".format(
                    args.resume, checkpoint["epoch"]
                )
            )
        else:
            print("=> no checkpoint found at '{}'".format(args.resume))

    if args.distributed:
        model = model.with_bagua(
            [optimizer],
            algorithm,
        )

    cudnn.benchmark = True

    # Data loading code
    traindir = os.path.join(args.data, "train")
    valdir = os.path.join(args.data, "val")
    normalize = transforms.Normalize(
        mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
    )

    train_dataset = datasets.ImageFolder(
        traindir,
        transforms.Compose(
            [
                transforms.RandomResizedCrop(224),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
                normalize,
            ]
        ),
    )
    val_dataset = datasets.ImageFolder(
        valdir,
        transforms.Compose(
            [
                transforms.Resize(256),
                transforms.CenterCrop(224),
                transforms.ToTensor(),
                normalize,
            ]
        ),
    )

    if args.distributed:
        train_sampler = torch.utils.data.distributed.DistributedSampler(train_dataset)
    else:
        train_sampler = None

    train_loader = torch.utils.data.DataLoader(
        train_dataset,
        batch_size=args.batch_size,
        shuffle=(train_sampler is None),
        num_workers=args.workers,
        pin_memory=True,
        sampler=train_sampler,
    )

    val_loader = torch.utils.data.DataLoader(
        val_dataset,
        batch_size=args.batch_size,
        shuffle=False,
        num_workers=args.workers,
        pin_memory=True,
    )

    if args.evaluate:
        validate(val_loader, model, criterion, 0, args)
        return

    for epoch in range(args.start_epoch, args.epochs):
        if args.distributed:
            train_sampler.set_epoch(epoch)

        if args.algorithm == "async":
            model.bagua_algorithm.resume(model)

        # train for one epoch
        train(train_loader, model, criterion, optimizer, scaler, epoch, args)

        if args.algorithm == "async":
            model.bagua_algorithm.abort(model)

        # evaluate on validation set
        acc1 = validate(val_loader, model, criterion, epoch, args)

        # remember best acc@1 and save checkpoint
        is_best = acc1 > best_acc1
        best_acc1 = max(acc1, best_acc1)

        if bagua.get_rank() == 0 and args.save_checkpoint:
            save_checkpoint(
                {
                    "epoch": epoch + 1,
                    "arch": args.arch,
                    "state_dict": model.state_dict(),
                    "best_acc1": best_acc1,
                    "optimizer": optimizer.state_dict(),
                },
                is_best,
            )


def train(train_loader, model, criterion, optimizer, scaler, epoch, args):
    batch_time = AverageMeter("Time", ":6.3f")
    data_time = AverageMeter("Data", ":6.3f")
    losses = AverageMeter("Loss", ":.4e")
    top1 = AverageMeter("Acc@1", ":6.2f")
    top5 = AverageMeter("Acc@5", ":6.2f")
    progress = ProgressMeter(
        len(train_loader),
        [batch_time, data_time, losses, top1, top5],
        prefix="Epoch: [{}]".format(epoch),
    )

    # switch to train mode
    model.train()

    end = time.time()
    for i, (images, target) in enumerate(train_loader):

        if args.prof >= 0 and i == args.prof:
            print("Profiling begun at iteration {}".format(i))
            torch.cuda.cudart().cudaProfilerStart()

        if args.prof >= 0:
            torch.cuda.nvtx.range_push("Body of iteration {}".format(i))

        # measure data loading time
        data_time.update(time.time() - end)

        if torch.cuda.is_available():
            images = images.cuda(bagua.get_local_rank(), non_blocking=True)
            target = target.cuda(bagua.get_local_rank(), non_blocking=True)

        adjust_learning_rate(optimizer, epoch, i, len(train_loader), args)

        optimizer.zero_grad()

        if args.prof >= 0:
            torch.cuda.nvtx.range_push("forward")

        with torch.cuda.amp.autocast(enabled=args.amp):
            # compute output
            output = model(images)
            loss = criterion(output, target)

        if args.prof >= 0:
            torch.cuda.nvtx.range_pop()

        # measure accuracy and record loss
        acc1, acc5 = accuracy(output, target, topk=(1, 5))
        losses.update(loss.item(), images.size(0))
        top1.update(acc1[0], images.size(0))
        top5.update(acc5[0], images.size(0))

        if args.prof >= 0:
            torch.cuda.nvtx.range_push("backward")

        # compute gradient and do SGD step
        scaler.scale(loss).backward()

        if args.prof >= 0:
            torch.cuda.nvtx.range_pop()

        if args.prof >= 0:
            torch.cuda.nvtx.range_push("optimizer.step()")

        scaler.step(optimizer)
        scaler.update()

        if args.prof >= 0:
            torch.cuda.nvtx.range_pop()

        # measure elapsed time
        batch_time.update(time.time() - end)
        end = time.time()

        if i % args.print_freq == 0:
            progress.display(i)

        # Pop range "Body of iteration {}".format(i)
        if args.prof >= 0:
            torch.cuda.nvtx.range_pop()

        if args.prof >= 0 and i == args.prof + 10:
            print("Profiling ended at iteration {}".format(i))
            torch.cuda.cudart().cudaProfilerStop()

            if args.algorithm == "async":
                model.bagua_algorithm.abort(model)
            quit()


def validate(val_loader, model, criterion, epoch, args):
    batch_time = AverageMeter("Time", ":6.3f")
    losses = AverageMeter("Loss", ":.4e")
    top1 = AverageMeter("Acc@1", ":6.2f")
    top5 = AverageMeter("Acc@5", ":6.2f")
    progress = ProgressMeter(
        len(val_loader), [batch_time, losses, top1, top5], prefix="Test: "
    )

    # switch to evaluate mode
    model.eval()

    with torch.no_grad():
        end = time.time()
        for i, (images, target) in enumerate(val_loader):
            if torch.cuda.is_available():
                images = images.cuda(bagua.get_local_rank(), non_blocking=True)
                target = target.cuda(bagua.get_local_rank(), non_blocking=True)

            # compute output
            output = model(images)
            loss = criterion(output, target)

            # measure accuracy and record loss
            acc1, acc5 = accuracy(output, target, topk=(1, 5))
            losses.update(loss.item(), images.size(0))
            top1.update(acc1[0], images.size(0))
            top5.update(acc5[0], images.size(0))

            # measure elapsed time
            batch_time.update(time.time() - end)
            end = time.time()

            if i % args.print_freq == 0:
                progress.display(i)

        # TODO: this should also be done with the ProgressMeter
        logging.info(
            " * TEST Epoch {} Acc@1 {top1.avg:.3f} Acc@5 {top5.avg:.3f}".format(
                epoch, top1=top1, top5=top5
            )
        )

    return top1.avg


def save_checkpoint(state, is_best, filename="checkpoint.pth.tar"):
    torch.save(state, filename)
    if is_best:
        shutil.copyfile(filename, "model_best.pth.tar")


class AverageMeter(object):
    """Computes and stores the average and current value"""

    def __init__(self, name, fmt=":f"):
        self.name = name
        self.fmt = fmt
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def __str__(self):
        fmtstr = "{name} {val" + self.fmt + "} ({avg" + self.fmt + "})"
        return fmtstr.format(**self.__dict__)


class ProgressMeter(object):
    def __init__(self, num_batches, meters, prefix=""):
        self.batch_fmtstr = self._get_batch_fmtstr(num_batches)
        self.meters = meters
        self.prefix = prefix

    def display(self, batch):
        entries = [self.prefix + self.batch_fmtstr.format(batch)]
        entries += [str(meter) for meter in self.meters]
        logging.info("\t".join(entries))

    def _get_batch_fmtstr(self, num_batches):
        num_digits = len(str(num_batches // 1))
        fmt = "{:" + str(num_digits) + "d}"
        return "[" + fmt + "/" + fmt.format(num_batches) + "]"


def adjust_learning_rate(optimizer, epoch, step, len_epoch, args):
    """Sets the learning rate to the initial LR decayed by 10 every 30 epochs"""
    # lr = args.lr * (0.1 ** (epoch // 30))
    # for param_group in optimizer.param_groups:
    #     param_group["lr"] = lr
    milestones = [int(i) for i in args.milestones.split(",")]
    lr = args.lr * (args.gama ** bisect_right(milestones, epoch))

    """Warmup"""
    if epoch < args.warmup_epochs:
        lr = (
            lr
            * float(1 + step + epoch * len_epoch)
            / float(args.warmup_epochs * len_epoch)
        )

    # logging.info("epoch = {}, step = {}, lr = {}".format(epoch, step, lr))

    for param_group in optimizer.param_groups:
        param_group["lr"] = lr


def accuracy(output, target, topk=(1,)):
    """Computes the accuracy over the k top predictions for the specified values of k"""
    with torch.no_grad():
        maxk = max(topk)
        batch_size = target.size(0)

        _, pred = output.topk(maxk, 1, True, True)
        pred = pred.t()
        correct = pred.eq(target.view(1, -1).expand_as(pred))

        res = []
        for k in topk:
            correct_k = correct[:k].reshape(-1).float().sum(0, keepdim=True)
            res.append(correct_k.mul_(100.0 / batch_size))
        return res


if __name__ == "__main__":
    main()



import os
import unittest

import torch
import torch.distributed as c10d

import bagua.torch_api as bagua
from tests.internal.common_utils import find_free_port
from tests.internal.multi_process import MultiProcessTestCase, setup_bagua_env
from tests import skip_if_cuda_not_available


class Result(object):
    def __init__(self):
        self.data = torch.zeros(100)


def run_new_group(rank, nprocs, args, results, env):
    setup_bagua_env(rank, env)

    all_ranks = list(range(nprocs))
    odd_ranks = list(filter(lambda r: r % 2 == 1, all_ranks))
    g = bagua.communication.new_group(ranks=odd_ranks)

    tensor = torch.rand(100).cuda()
    tensor *= rank

    bagua.communication.allreduce(tensor, tensor, comm=g.get_global_communicator())
    results[rank].data.copy_(tensor)


def run_from_torch_group(rank, nprocs, args, results, env):
    setup_bagua_env(rank, env)

    all_ranks = list(range(nprocs))
    ranks_1 = list(filter(lambda r: r % 3 == 1, all_ranks))
    ranks_2 = list(filter(lambda r: r % 2 == 0, all_ranks))

    g_1 = torch.distributed.new_group(ranks_1)
    bg_1 = bagua.communication.from_torch_group(g_1)

    g_2 = torch.distributed.new_group(ranks_2)
    bg_2 = bagua.communication.from_torch_group(g_2)

    if rank in ranks_1:
        assert torch.distributed.get_rank(g_1) == bg_1.get_global_communicator().rank()
        assert (
            torch.distributed.get_world_size(g_1)
            == bg_1.get_global_communicator().nranks()  # noqa: W503
        )

    if rank in ranks_2:
        assert torch.distributed.get_rank(g_2) == bg_2.get_global_communicator().rank()
        assert (
            torch.distributed.get_world_size(g_2)
            == bg_2.get_global_communicator().nranks()  # noqa: W503
        )


class TestProcessGroup(MultiProcessTestCase):
    @skip_if_cuda_not_available()
    def test_new_group(self):
        nprocs = torch.cuda.device_count()
        results = [Result() for _ in range(nprocs)]
        self.run_test_locally(run_new_group, nprocs, args={}, results=results)

        all_ranks = list(range(nprocs))
        odd_ranks = list(filter(lambda r: r % 2 == 1, all_ranks))

        for rank in odd_ranks:
            peer_rank = (rank + 2) % nprocs

            self.assertTrue(torch.equal(results[rank].data, results[peer_rank].data))

    @skip_if_cuda_not_available()
    def test_from_torch_group(self):
        nprocs = torch.cuda.device_count()
        self.run_test_locally(run_from_torch_group, nprocs, args={}, results=None)


from tests.internal.torch.common_distributed import (  # noqa: E402
    MultiProcessTestCase,
    skip_if_lt_x_gpu,
)


class ProcessGroupNCCLTest(MultiProcessTestCase):
    def setUp(self):
        super(ProcessGroupNCCLTest, self).setUp()
        # NCCL_BLOCKING_WAIT overrides NCCL_ASYNC_ERROR_HANDLING hence tests
        # that use NCCL_BLOCKING_WAIT will test it as expected.
        os.environ["NCCL_ASYNC_ERROR_HANDLING"] = "1"
        os.environ.update(
            {
                "MASTER_ADDR": "127.0.0.1",
                "MASTER_PORT": str(find_free_port(8000, 8100)),
                "BAGUA_SERVICE_PORT": str(find_free_port(9000, 9100)),
            }
        )
        self._spawn_processes()

    @skip_if_lt_x_gpu(2)
    def test_bagua_pg(self):
        # Need to use NCCL_BLOCKING_WAIT and not ASYNC_ERROR_HANDLING,
        # otherwise process will be taken down and we can't check for errors.
        os.environ["NCCL_ASYNC_ERROR_HANDLING"] = "0"
        os.environ["NCCL_BLOCKING_WAIT"] = "1"
        os.environ.update(
            {
                "WORLD_SIZE": str(self.world_size),
                "LOCAL_WORLD_SIZE": str(self.world_size),
                "RANK": str(self.rank),
                "LOCAL_RANK": str(self.rank),
            }
        )

        bagua.init_process_group()
        pg = c10d.new_group(ranks=list(range(0, self.world_size)))
        pg.bagua_patch()
        self.assertTrue(pg in bagua.communication._torch_to_bagua_pg_map)
        del pg
        c10d.destroy_process_group()
        self.assertEqual(len(bagua.communication._torch_to_bagua_pg_map), 0)


if __name__ == "__main__":
    unittest.main()



import torch
import torch.nn as nn
import torch.nn.functional as F
from tests.internal.common_utils import find_free_port
import unittest
import multiprocessing
import os
import bagua.torch_api as bagua
from tests import skip_if_cuda_not_available
import logging


class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.fc1 = nn.Linear(2, 10, bias=False)
        self.fc2 = nn.Linear(10, 50, bias=True)
        self.fc3 = nn.Linear(50, 4, bias=False)
        self.relu = nn.ReLU()

    def forward(self, x):
        x = self.relu(self.fc1(x))
        x = self.relu(self.fc2(x))
        x = self.fc3(x)
        return F.softmax(x, dim=1)


def run_model_wrapper(rank, env, fn, warmup_steps):
    # initialize subprocess env
    os.environ["WORLD_SIZE"] = env["WORLD_SIZE"]
    os.environ["LOCAL_WORLD_SIZE"] = env["LOCAL_WORLD_SIZE"]
    os.environ["MASTER_ADDR"] = env["MASTER_ADDR"]
    os.environ["MASTER_PORT"] = env["MASTER_PORT"]
    os.environ["BAGUA_SERVICE_PORT"] = env["BAGUA_SERVICE_PORT"]
    os.environ["RANK"] = str(rank)
    os.environ["LOCAL_RANK"] = str(rank)

    # init bagua distributed process group
    torch.cuda.set_device(rank)
    bagua.init_process_group()

    # construct model and optimizer, etc.
    model = Net().cuda()
    optimizer = torch.optim.SGD(model.parameters(), lr=0.01)
    loss_fn = nn.MSELoss()

    # wrap model
    algorithm = bagua.algorithms.async_model_average.AsyncModelAverageAlgorithm(
        sync_interval_ms=20,
        warmup_steps=warmup_steps,
    )
    model = model.with_bagua([optimizer], algorithm)

    fn(model, optimizer, loss_fn)


def train_epoch(epoch, model, optimizer, loss_fn):
    logging.debug("Training epoch {}".format(epoch))
    for _ in range(10):
        data = torch.randn(4, 2).cuda()
        target = torch.randn(4, 4).cuda()

        optimizer.zero_grad()
        output = model(data)
        loss = loss_fn(output, target)

        loss.backward()
        optimizer.step()


def run_epochs(model, optimizer, loss_fn):
    for epoch in range(100):
        train_epoch(epoch, model, optimizer, loss_fn)
    model.bagua_algorithm.abort(model)


def run_multiple_aborts(model, optimizer, loss_fn):
    for epoch in range(2):
        model.bagua_algorithm.resume(model)
        model.bagua_algorithm.abort(model)
        model.bagua_algorithm.resume(model)
        for _ in range(100):
            train_epoch(epoch, model, optimizer, loss_fn)

        model.bagua_algorithm.abort(model)
        model.bagua_algorithm.abort(model)


class TestAsyncModelAverage(unittest.TestCase):
    @skip_if_cuda_not_available()
    def test_algorithm(self):
        nprocs = torch.cuda.device_count()
        env = {
            "WORLD_SIZE": str(nprocs),
            "LOCAL_WORLD_SIZE": str(nprocs),
            "MASTER_ADDR": "127.0.0.1",
            "MASTER_PORT": str(find_free_port(8000, 8100)),
            "BAGUA_SERVICE_PORT": str(find_free_port(9000, 9100)),
        }

        mp = multiprocessing.get_context("spawn")
        processes = []
        for i in range(nprocs):
            p = mp.Process(target=run_model_wrapper, args=(i, env, run_epochs, 0))
            p.start()
            processes.append(p)

        for p in processes:
            p.join(timeout=60)
            self.assertTrue(p.exitcode == 0)

    @skip_if_cuda_not_available()
    def test_multiple_aborts(self):
        nprocs = torch.cuda.device_count()
        env = {
            "WORLD_SIZE": str(nprocs),
            "LOCAL_WORLD_SIZE": str(nprocs),
            "MASTER_ADDR": "127.0.0.1",
            "MASTER_PORT": str(find_free_port(8000, 8100)),
            "BAGUA_SERVICE_PORT": str(find_free_port(9000, 9100)),
        }

        mp = multiprocessing.get_context("spawn")
        processes = []
        for i in range(nprocs):
            p = mp.Process(
                target=run_model_wrapper, args=(i, env, run_multiple_aborts, 10)
            )
            p.start()
            processes.append(p)

        for p in processes:
            p.join(timeout=60)
            self.assertTrue(p.exitcode == 0)


if __name__ == "__main__":
    unittest.main()



import unittest
from bagua.torch_api.contrib.utils.redis_store import (
    RedisStore,
    start_redis_server_cli,
)
from bagua.torch_api.env import find_free_network_port
import redis
import multiprocessing as mp
import logging
import numpy as np
import pickle
from tests import skip_if_cuda_available


logging.basicConfig(level=logging.DEBUG)


class TestRedisStore(unittest.TestCase):
    def check(self, store):
        store.clear()
        self.assertEqual(store.num_keys(), 0)

        self.generated_data = [np.random.rand(10) for _ in range(5)]
        store.set("1", pickle.dumps(self.generated_data[1]))

        store.mset(
            {
                "2": pickle.dumps(self.generated_data[2]),
                "3": pickle.dumps(self.generated_data[3]),
                "4": pickle.dumps(self.generated_data[4]),
            }
        )
        ret = store.mget(["1", "2", "5"])
        self.assertTrue((pickle.loads(ret[0]) == self.generated_data[1]).all())
        self.assertTrue((pickle.loads(ret[1]) == self.generated_data[2]).all())
        self.assertEqual(ret[2], None)

        r1 = store.get("4")
        r2 = store.get("6")
        self.assertTrue((pickle.loads(r1) == self.generated_data[4]).all())
        self.assertEqual(r2, None)

        cnt = store.num_keys()
        self.assertEqual(cnt, 4)

        self.assertTrue(store.status())

    @skip_if_cuda_available()
    def test_redis_store(self):
        store = RedisStore(hosts=None, cluster_mode=False, capacity_per_node=10000000)
        self.check(store)

    @skip_if_cuda_available()
    def test_redis_cluster_store(self):
        n = 3
        hosts = []
        ports = []
        processes = []
        for i in range(n):
            port = find_free_network_port()
            p = mp.Process(
                target=start_redis_server_cli,
                args=(port, 10000000, f"--cluster-config-file nodes{port}.conf"),
            )
            p.start()

            ports.append(port)
            processes.append(p)
            hosts.append({"host": "127.0.0.1", "port": port})

        for p in processes:
            p.join()
            self.assertTrue(p.exitcode == 0)

        store = RedisStore(hosts=hosts, cluster_mode=True)
        self.check(store)

        # Shut down servers manually
        for port in ports:
            client = redis.Redis(port=port)
            client.shutdown(nosave=True)
            client.close()


if __name__ == "__main__":
    unittest.main()



"""
This script is modified from the horovod benchmark script(https://github.com/horovod/horovod/blob/master/examples/pytorch/pytorch_synthetic_benchmark.py).
You can compare the performance with the horovod script directly.
"""

import argparse
import timeit
import torch.backends.cudnn as cudnn
import torch.nn.functional as F
import torch.optim as optim
import torch.utils.data.distributed
from torchvision import models
import numpy as np
import bagua.torch_api as bagua
import logging


logging.basicConfig(format="%(asctime)s %(levelname)s:%(message)s", level=logging.ERROR)
if bagua.get_rank() == 0:
    logging.getLogger().setLevel(logging.INFO)

# Benchmark settings
parser = argparse.ArgumentParser(
    description="PyTorch Synthetic Benchmark",
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
)

parser.add_argument("--model", type=str, default="resnet50", help="model to benchmark")
parser.add_argument("--batch-size", type=int, default=32, help="input batch size")

parser.add_argument(
    "--num-warmup-batches",
    type=int,
    default=10,
    help="number of warm-up batches that don't count towards benchmark",
)
parser.add_argument(
    "--num-batches-per-iter",
    type=int,
    default=10,
    help="number of batches per benchmark iteration",
)
parser.add_argument(
    "--num-iters", type=int, default=10, help="number of benchmark iterations"
)

parser.add_argument(
    "--no-cuda", action="store_true", default=False, help="disables CUDA training"
)

parser.add_argument(
    "--deterministic",
    action="store_true",
    default=False,
    help="deterministic reproducible training",
)

# bagua args
parser.add_argument(
    "--algorithm",
    type=str,
    default="gradient_allreduce",
    help="gradient_allreduce, bytegrad, decentralized, low_precision_decentralized, qadam or async",
)
parser.add_argument(
    "--async-sync-interval",
    default=500,
    type=int,
    help="Model synchronization interval(ms) for async algorithm",
)
parser.add_argument(
    "--async-warmup-steps",
    default=0,
    type=int,
    help="Warmup(allreduce) steps for async algorithm",
)
parser.add_argument(
    "--amp",
    action="store_true",
    default=False,
    help="use amp",
)
parser.add_argument(
    "--log-interval",
    type=int,
    default=10,
    metavar="N",
    help="how many batches to wait before logging training status",
)
parser.add_argument(
    "--fuse-optimizer",
    action="store_true",
    default=False,
    help="fuse optimizer or not",
)

args = parser.parse_args()
args.cuda = not args.no_cuda and torch.cuda.is_available()

if not args.cuda:
    raise RuntimeError("bagua currently not supporting non-GPU mode")

if args.cuda:
    torch.cuda.set_device(bagua.get_local_rank())
bagua.init_process_group()

scaler = torch.cuda.amp.GradScaler(enabled=args.amp)

cudnn.benchmark = True
if args.deterministic:
    cudnn.benchmark = False
    cudnn.deterministic = True
    np.random.seed(bagua.get_rank())
    torch.manual_seed(bagua.get_rank())
    torch.cuda.manual_seed(bagua.get_rank())


# Set up standard model.
model = getattr(models, args.model)()

if args.cuda:
    # Move model to GPU.
    model.cuda()

optimizer = optim.SGD(model.parameters(), lr=0.01 * bagua.get_world_size())

if args.algorithm == "gradient_allreduce":
    from bagua.torch_api.algorithms import gradient_allreduce

    algorithm = gradient_allreduce.GradientAllReduceAlgorithm()
elif args.algorithm == "decentralized":
    from bagua.torch_api.algorithms import decentralized

    algorithm = decentralized.DecentralizedAlgorithm()
elif args.algorithm == "low_precision_decentralized":
    from bagua.torch_api.algorithms import decentralized

    algorithm = decentralized.LowPrecisionDecentralizedAlgorithm()
elif args.algorithm == "bytegrad":
    from bagua.torch_api.algorithms import bytegrad

    algorithm = bytegrad.ByteGradAlgorithm()
elif args.algorithm == "qadam":
    from bagua.torch_api.algorithms import q_adam

    optimizer = q_adam.QAdamOptimizer(
        model.parameters(), lr=0.01 * bagua.get_world_size(), warmup_steps=100
    )
    algorithm = q_adam.QAdamAlgorithm(optimizer)
elif args.algorithm == "async":
    from bagua.torch_api.algorithms import async_model_average

    algorithm = async_model_average.AsyncModelAverageAlgorithm(
        sync_interval_ms=args.async_sync_interval,
        warmup_steps=args.async_warmup_steps,
    )
else:
    raise NotImplementedError

model = model.with_bagua([optimizer], algorithm, do_flatten=not args.fuse_optimizer)

if args.fuse_optimizer:
    optimizer = bagua.contrib.fuse_optimizer(optimizer)

# Set up fixed fake data
data = torch.randn(args.batch_size, 3, 224, 224)
target = torch.LongTensor(args.batch_size).random_() % 1000
if args.cuda:
    data, target = data.cuda(), target.cuda()


batch_idx = 0


def benchmark_step():
    global batch_idx
    optimizer.zero_grad()
    output = model(data)
    loss = F.cross_entropy(output, target)

    if args.amp:
        scaled_loss = args.scaler.scale(loss)
        scaled_loss.backward()
        scaler.step(optimizer)
        scaler.update()
        return

    loss.backward()
    optimizer.step()
    if batch_idx % args.log_interval == 0:
        logging.info(
            "BatchIdx: {} TrainLoss: {:.6f}".format(
                batch_idx,
                loss.item(),
            )
        )
    batch_idx += 1


logging.info("Model: %s" % args.model)
logging.info("Batch size: %d" % args.batch_size)
device = "GPU" if args.cuda else "CPU"
logging.info("Number of %ss: %d" % (device, bagua.get_world_size()))

# Warm-up
logging.info("Running warmup...")

timeit.timeit(benchmark_step, number=args.num_warmup_batches)

# Benchmark
logging.info("Running benchmark...")
img_secs = []
for x in range(args.num_iters):
    time = timeit.timeit(benchmark_step, number=args.num_batches_per_iter)
    img_sec = args.batch_size * args.num_batches_per_iter / time
    logging.info(
        "Iter #%d: %.1f img/sec %s" % (x, img_sec * bagua.get_world_size(), device)
    )
    img_secs.append(img_sec)

# Results
img_sec_mean = np.mean(img_secs)
img_sec_conf = 1.96 * np.std(img_secs)
logging.info("Img/sec per %s: %.1f +-%.1f" % (device, img_sec_mean, img_sec_conf))
logging.info(
    "Total img/sec on %d %s(s): %.1f +-%.1f"
    % (
        bagua.get_world_size(),
        device,
        bagua.get_world_size() * img_sec_mean,
        bagua.get_world_size() * img_sec_conf,
    )
)

if args.algorithm == "async":
    model.bagua_algorithm.abort(model)



import torch
import torch.nn as nn
import torch.nn.functional as F
from tests.internal.common_utils import find_free_port
import unittest
import multiprocessing
import os
from bagua.torch_api.utils import flatten
import bagua.torch_api as bagua
from tests import skip_if_cuda_not_available
from bagua.torch_api.data_parallel import DistributedDataParallel as DDP

N_EPOCHS = 10


class Net1(nn.Module):
    def __init__(self):
        super(Net1, self).__init__()
        self.fc1 = nn.Linear(2, 10, bias=False)
        self.fc2 = nn.Linear(10, 50, bias=True)
        self.fc3 = nn.Linear(50, 4, bias=False)
        self.relu = nn.ReLU()

    def forward(self, x):
        x = self.relu(self.fc1(x))
        x = self.relu(self.fc2(x))
        x = self.fc3(x)
        return F.softmax(x, dim=1)


class Net2(nn.Module):
    def __init__(self):
        super(Net2, self).__init__()
        self.fc1 = nn.Linear(2, 10, bias=False)
        self.fc2 = nn.Linear(10, 30, bias=True)
        self.fc3 = nn.Linear(30, 20, bias=True)
        self.fc4 = nn.Linear(20, 4, bias=False)
        self.relu = nn.ReLU()

    def forward(self, x):
        x = self.relu(self.fc1(x))
        x = self.relu(self.fc2(x))
        x = self.relu(self.fc3(x))
        x = self.fc4(x)
        return F.softmax(x, dim=1)


def _init_bagua_env(rank, env):
    # set deterministic
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True
    torch.manual_seed(rank)
    # initialize subprocess env
    os.environ["WORLD_SIZE"] = env["WORLD_SIZE"]
    os.environ["LOCAL_WORLD_SIZE"] = env["LOCAL_WORLD_SIZE"]
    os.environ["MASTER_ADDR"] = env["MASTER_ADDR"]
    os.environ["MASTER_PORT"] = env["MASTER_PORT"]
    os.environ["BAGUA_SERVICE_PORT"] = env["BAGUA_SERVICE_PORT"]

    os.environ["RANK"] = str(rank)
    os.environ["LOCAL_RANK"] = str(rank)

    # init bagua distributed process group
    torch.cuda.set_device(rank)
    bagua.init_process_group()


def _init_torch_env(rank, nprocs, backend):
    # set deterministic
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True
    torch.manual_seed(rank)

    # init torch distributed process group
    torch.cuda.set_device(rank)
    torch.distributed.init_process_group(
        world_size=nprocs,
        rank=rank,
        backend=backend,
        init_method="file:///tmp/.bagua.test.filestore",
    )


def run_model(
    rank,
    results,
    env,
):
    _init_bagua_env(rank, env)

    # construct model and optimizer, etc.
    model_1 = Net1().cuda()
    optimizer_1 = torch.optim.SGD(model_1.parameters(), lr=0.01)
    loss_fn_1 = nn.MSELoss()

    model_2 = Net2().cuda()
    optimizer_2 = torch.optim.SGD(model_2.parameters(), lr=0.01)
    loss_fn_2 = nn.MSELoss()

    # wrap model
    from bagua.torch_api.algorithms import gradient_allreduce

    algorithm = gradient_allreduce.GradientAllReduceAlgorithm()

    model_1 = DDP(model_1, optimizers=[optimizer_1], algorithm=algorithm)

    model_2 = DDP(model_2, optimizers=[optimizer_2], algorithm=algorithm)

    ret = results[rank]

    ret.init_weight_1.copy_(flatten([param.data for param in model_1.parameters()]))
    ret.init_weight_2.copy_(flatten([param.data for param in model_2.parameters()]))

    for epoch in range(N_EPOCHS):
        data_1 = torch.randn(8, 2).cuda()
        target_1 = torch.randn(8, 4).cuda()

        optimizer_1.zero_grad()
        output_1 = model_1(data_1)
        loss_1 = loss_fn_1(output_1, target_1)

        loss_1.backward()
        optimizer_1.step()

        data_2 = torch.randn(8, 2).cuda()
        target_2 = torch.randn(8, 4).cuda()

        optimizer_2.zero_grad()
        output_2 = model_2(data_2)
        loss_2 = loss_fn_2(output_2, target_2)

        loss_2.backward()
        optimizer_2.step()

    ret.end_weight_1.copy_(flatten([param.data for param in model_1.parameters()]))
    ret.end_weight_2.copy_(flatten([param.data for param in model_2.parameters()]))


def run_torch_model(
    rank,
    nprocs,
    results,
    backend,
    env,
):
    _init_torch_env(rank, nprocs, backend)

    # construct model and optimizer, etc.
    model_1 = Net1().cuda()
    optimizer_1 = torch.optim.SGD(model_1.parameters(), lr=0.01)
    loss_fn_1 = nn.MSELoss()

    model_2 = Net2().cuda()
    optimizer_2 = torch.optim.SGD(model_2.parameters(), lr=0.01)
    loss_fn_2 = nn.MSELoss()

    # wrap model
    model_1 = torch.nn.parallel.DistributedDataParallel(model_1, device_ids=[rank])
    model_2 = torch.nn.parallel.DistributedDataParallel(model_2, device_ids=[rank])

    ret = results[rank]

    ret.init_weight_1.copy_(flatten([param.data for param in model_1.parameters()]))
    ret.init_weight_2.copy_(flatten([param.data for param in model_2.parameters()]))

    for epoch in range(N_EPOCHS):
        data_1 = torch.randn(8, 2).cuda()
        target_1 = torch.randn(8, 4).cuda()

        optimizer_1.zero_grad()
        output_1 = model_1(data_1)
        loss_1 = loss_fn_1(output_1, target_1)

        loss_1.backward()
        optimizer_1.step()

        data_2 = torch.randn(8, 2).cuda()
        target_2 = torch.randn(8, 4).cuda()

        optimizer_2.zero_grad()
        output_2 = model_2(data_2)
        loss_2 = loss_fn_2(output_2, target_2)

        loss_2.backward()
        optimizer_2.step()

    ret.end_weight_1.copy_(flatten([param.data for param in model_1.parameters()]))
    ret.end_weight_2.copy_(flatten([param.data for param in model_2.parameters()]))


class Result(object):
    def __init__(self):
        model_1 = Net1()
        model_2 = Net2()
        self.init_weight_1 = flatten(
            [torch.zeros_like(param.data) for param in model_1.parameters()]
        )
        self.end_weight_1 = flatten(
            [torch.zeros_like(param.data) for param in model_1.parameters()]
        )

        self.init_weight_2 = flatten(
            [torch.zeros_like(param.data) for param in model_2.parameters()]
        )
        self.end_weight_2 = flatten(
            [torch.zeros_like(param.data) for param in model_2.parameters()]
        )


class TestMultiModels(unittest.TestCase):
    @skip_if_cuda_not_available()
    def test_multi_models(self):
        nprocs = torch.cuda.device_count()
        env = {}
        mp = multiprocessing.get_context("spawn")
        torch_results = [Result() for _ in range(nprocs)]
        processes = []
        backend = "gloo"
        for i in range(nprocs):
            p = mp.Process(
                target=run_torch_model,
                args=(
                    i,
                    nprocs,
                    torch_results,
                    backend,
                    env,
                ),
            )
            p.start()
            processes.append(p)

        for p in processes:
            p.join(timeout=60)
            self.assertTrue(p.exitcode == 0)

        env = {
            "WORLD_SIZE": str(nprocs),
            "LOCAL_WORLD_SIZE": str(nprocs),
            "MASTER_ADDR": "127.0.0.1",
            "MASTER_PORT": str(find_free_port(8000, 8100)),
            "BAGUA_SERVICE_PORT": str(find_free_port(9000, 9100)),
        }

        bagua_results = [Result() for _ in range(nprocs)]
        processes = []
        for i in range(nprocs):
            p = mp.Process(
                target=run_model,
                args=(
                    i,
                    bagua_results,
                    env,
                ),
            )
            p.start()
            processes.append(p)

        for p in processes:
            p.join(timeout=60)
            self.assertTrue(p.exitcode == 0)

        for rank in range(nprocs):
            self.assertTrue(
                torch.all(
                    torch.isclose(
                        bagua_results[rank].init_weight_1,
                        torch_results[rank].init_weight_1,
                    )
                ).item()
            )

            self.assertTrue(
                torch.all(
                    torch.isclose(
                        bagua_results[rank].end_weight_1,
                        torch_results[rank].end_weight_1,
                    )
                ).item()
            )
            self.assertTrue(
                torch.all(
                    torch.isclose(
                        bagua_results[rank].init_weight_2,
                        torch_results[rank].init_weight_2,
                    )
                ).item()
            )

            self.assertTrue(
                torch.all(
                    torch.isclose(
                        bagua_results[rank].end_weight_2,
                        torch_results[rank].end_weight_2,
                    )
                ).item()
            )


if __name__ == "__main__":
    unittest.main()



import os
import unittest
import multiprocessing

import bagua.torch_api as bagua
from bagua.torch_api.contrib.sync_batchnorm import SyncBatchNorm

import torch

from tests.internal.common_utils import find_free_port


class Result_Forward(object):
    def __init__(self):
        bn_layer = SyncBatchNorm(num_features=8)
        self.bn_output = torch.zeros(32, 8, 16, 16).cuda().float()
        self.running_mean = torch.empty_like(bn_layer.running_mean)
        self.running_var = torch.empty_like(bn_layer.running_var)


class Result_Backward(object):
    def __init__(self):
        bn_layer = SyncBatchNorm(num_features=8)
        self.weight_grad = torch.empty_like(bn_layer.weight.data)
        self.bias_grad = torch.empty_like(bn_layer.bias.data)
        self.input_grad = torch.zeros(32, 8, 16, 16).cuda().float()


def _init_torch_env(rank, nprocs, env):
    # set deterministic
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True
    torch.manual_seed(rank)

    # init torch distributed process group
    torch.cuda.set_device(rank)
    init_method = "tcp://" + env["MASTER_ADDR"] + ":" + env["MASTER_PORT"]
    torch.distributed.init_process_group(
        world_size=nprocs,
        rank=rank,
        backend="nccl",
        init_method=init_method,
    )


def _init_bagua_env(rank, env):
    # set deterministic
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True
    torch.manual_seed(rank)
    # initialize subprocess env
    os.environ["WORLD_SIZE"] = env["WORLD_SIZE"]
    os.environ["LOCAL_WORLD_SIZE"] = env["LOCAL_WORLD_SIZE"]
    os.environ["MASTER_ADDR"] = env["MASTER_ADDR"]
    os.environ["MASTER_PORT"] = env["MASTER_PORT"]
    os.environ["BAGUA_SERVICE_PORT"] = env["BAGUA_SERVICE_PORT"]

    os.environ["RANK"] = str(rank)
    os.environ["LOCAL_RANK"] = str(rank)

    # init bagua distributed process group
    torch.cuda.set_device(rank)
    bagua.init_process_group()


def run_torch_sync_bn(rank, nprocs, torch_results_forward, results_backward, envs):
    _init_torch_env(rank, nprocs, envs)

    input_d = torch.randn(32, 8, 16, 16).cuda().float()
    input_d.requires_grad_()
    sync_bn_ddp = torch.nn.SyncBatchNorm(num_features=8)

    sync_bn_ddp.cuda(rank)
    input_d = input_d.cuda(rank)
    sync_bn_ddp_out = sync_bn_ddp(input_d)

    ddp_ret1 = torch_results_forward[rank]
    ddp_ret1.bn_output.copy_(sync_bn_ddp_out)
    ddp_ret1.running_mean.copy_(sync_bn_ddp.running_mean)
    ddp_ret1.running_var.copy_(sync_bn_ddp.running_var)

    sync_bn_ddp_out.sum().backward()

    ddp_ret2 = results_backward[rank]
    ddp_ret2.weight_grad.copy_(sync_bn_ddp.weight.grad)
    ddp_ret2.bias_grad.copy_(sync_bn_ddp.bias.grad)
    ddp_ret2.input_grad.copy_(input_d.grad)


def run_bagua_sync_bn(rank, nprocs, bagua_results_forward, results_backward, envs):
    _init_bagua_env(rank, envs)

    input_b = torch.randn(32, 8, 16, 16).cuda().float()
    input_b.requires_grad_()
    sync_bn_bagua = SyncBatchNorm(num_features=8)

    sync_bn_bagua.cuda(rank)
    input_b = input_b.cuda(rank)
    sync_bn_bagua_out = sync_bn_bagua(input_b)

    bagua_ret1 = bagua_results_forward[rank]
    bagua_ret1.bn_output.copy_(sync_bn_bagua_out)
    bagua_ret1.running_mean.copy_(sync_bn_bagua.running_mean)
    bagua_ret1.running_var.copy_(sync_bn_bagua.running_var)

    sync_bn_bagua_out.sum().backward()

    bagua_ret2 = results_backward[rank]
    bagua_ret2.weight_grad.copy_(sync_bn_bagua.weight.grad)
    bagua_ret2.bias_grad.copy_(sync_bn_bagua.bias.grad)
    bagua_ret2.input_grad.copy_(input_b.grad)


class Test_Sync_Bn(unittest.TestCase):
    def test_syncbn(self):
        nprocs = torch.cuda.device_count()
        print(nprocs)
        env = {
            "MASTER_ADDR": "127.0.0.1",
            "MASTER_PORT": str(find_free_port(8000, 8100)),
        }

        mp = multiprocessing.get_context("spawn")
        torch_results_forward = [Result_Forward() for _ in range(nprocs)]
        torch_results_backward = [Result_Backward() for _ in range(nprocs)]
        processes = []

        for i in range(nprocs):
            p = mp.Process(
                target=run_torch_sync_bn,
                args=(
                    i,
                    nprocs,
                    torch_results_forward,
                    torch_results_backward,
                    env,
                ),
            )
            p.start()
            processes.append(p)

        for p in processes:
            p.join()
            self.assertTrue(p.exitcode == 0)

        env = {
            "WORLD_SIZE": str(nprocs),
            "LOCAL_WORLD_SIZE": str(nprocs),
            "MASTER_ADDR": "127.0.0.1",
            "MASTER_PORT": str(find_free_port(8000, 8100)),
            "BAGUA_SERVICE_PORT": str(find_free_port(9000, 9100)),
        }

        bagua_results_forward = [Result_Forward() for _ in range(nprocs)]
        bagua_results_backward = [Result_Backward() for _ in range(nprocs)]
        processes = []
        for i in range(nprocs):
            p = mp.Process(
                target=run_bagua_sync_bn,
                args=(
                    i,
                    nprocs,
                    bagua_results_forward,
                    bagua_results_backward,
                    env,
                ),
            )
            p.start()
            processes.append(p)
        for p in processes:
            p.join()
            self.assertTrue(p.exitcode == 0)

        for rank in range(nprocs):
            self.assertTrue(
                torch.all(
                    torch.isclose(
                        torch_results_forward[rank].bn_output,
                        bagua_results_forward[rank].bn_output,
                    )
                ).item()
            )
            self.assertTrue(
                torch.all(
                    torch.isclose(
                        torch_results_forward[rank].running_mean,
                        bagua_results_forward[rank].running_mean,
                    )
                ).item()
            )
            self.assertTrue(
                torch.all(
                    torch.isclose(
                        torch_results_forward[rank].running_var,
                        bagua_results_forward[rank].running_var,
                    )
                ).item()
            )

            self.assertTrue(
                torch.all(
                    torch.isclose(
                        torch_results_backward[rank].weight_grad,
                        bagua_results_backward[rank].weight_grad,
                    )
                ).item()
            )
            self.assertTrue(
                torch.all(
                    torch.isclose(
                        torch_results_backward[rank].bias_grad,
                        bagua_results_backward[rank].bias_grad,
                    )
                ).item()
            )
            self.assertTrue(
                torch.all(
                    torch.isclose(
                        torch_results_backward[rank].input_grad,
                        bagua_results_backward[rank].input_grad,
                    )
                ).item()
            )


if __name__ == "__main__":
    unittest.main()



import torch
import torch.nn as nn
import torch.nn.functional as F
from tests.internal.common_utils import find_free_port
from tests.internal.multi_process import setup_bagua_env
import unittest
import multiprocessing
from tests import skip_if_cuda_not_available
from bagua.torch_api.data_parallel import DistributedDataParallel as DDP


class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.fc1 = nn.Linear(2, 10, bias=False)
        self.fc2 = nn.Linear(10, 50, bias=True)
        self.fc3 = nn.Linear(50, 4, bias=False)
        self.relu = nn.ReLU()

    def forward(self, x):
        x = self.relu(self.fc1(x))
        x = self.relu(self.fc2(x))
        x = self.fc3(x)
        return F.softmax(x, dim=1)


def train(model, optimizer, loss_fn, is_async):
    if is_async:
        model.bagua_algorithm.resume(model)

    for _ in range(1000):
        data = torch.randn(4, 2).cuda()
        target = torch.randn(4, 4).cuda()

        optimizer.zero_grad()
        output = model(data)
        loss = loss_fn(output, target)

        loss.backward()
        optimizer.step()

    if is_async:
        model.bagua_algorithm.abort(model)


def bagua_init(model, optimizer, algorithm, process_group=None):
    # wrap model
    if algorithm == "gradient_allreduce":
        from bagua.torch_api.algorithms import gradient_allreduce

        bagua_algorithm = gradient_allreduce.GradientAllReduceAlgorithm()
    elif algorithm == "bytegrad":
        from bagua.torch_api.algorithms import bytegrad

        bagua_algorithm = bytegrad.ByteGradAlgorithm(hierarchical=False)
    elif algorithm == "decentralized":
        from bagua.torch_api.algorithms import decentralized

        bagua_algorithm = decentralized.DecentralizedAlgorithm(hierarchical=False)
    elif algorithm == "async":
        from bagua.torch_api.algorithms import async_model_average

        bagua_algorithm = async_model_average.AsyncModelAverageAlgorithm(
            sync_interval_ms=20,
        )
    elif algorithm == "low_prec_decentralized":
        from bagua.torch_api.algorithms import decentralized

        bagua_algorithm = decentralized.LowPrecisionDecentralizedAlgorithm(
            hierarchical=False
        )
    elif algorithm == "qadam":
        from bagua.torch_api.algorithms.q_adam import QAdamAlgorithm, QAdamOptimizer

        optimizer = QAdamOptimizer(model.parameters(), warmup_steps=10)
        bagua_algorithm = QAdamAlgorithm(optimizer, hierarchical=False)
    else:
        raise ValueError("unsupported algorithm")

    ddp_model = DDP(model, optimizers=[optimizer], algorithm=bagua_algorithm, process_group=process_group)

    return ddp_model, optimizer


def run_model_wrapper(rank, nranks, env, algorithm):
    # initialize subprocess env
    setup_bagua_env(rank, env)

    # construct model and optimizer, etc.
    model = Net().cuda()
    optimizer = torch.optim.SGD(model.parameters(), lr=0.01)
    loss_fn = nn.MSELoss()

    ddp_model, optimizer = bagua_init(
        model, optimizer, algorithm
    )

    train(ddp_model, optimizer, loss_fn, is_async=(algorithm == "async"))


def run_model_switch_wrapper(rank, nranks, env, algorithms):
    # initialize subprocess env
    setup_bagua_env(rank, env)

    # construct model and optimizer, etc.
    model = Net().cuda()
    optimizer = torch.optim.SGD(model.parameters(), lr=0.01)
    loss_fn = nn.MSELoss()

    for i in range(len(algorithms)):
        ddp_model, optimizer = bagua_init(model, optimizer, algorithms[i])
        train(ddp_model, optimizer, loss_fn, is_async=(algorithms[i] == "async"))
        print('algorithm={} done'.format(algorithms[i]))


class TestBaguaModule(unittest.TestCase):
    def run_algorithm(self, nprocs, nranks, fn, algorithm):
        env = {
            "WORLD_SIZE": str(nprocs),
            "LOCAL_WORLD_SIZE": str(nprocs),
            "MASTER_ADDR": "127.0.0.1",
            "MASTER_PORT": str(find_free_port(8000, 8100)),
            "BAGUA_SERVICE_PORT": str(find_free_port(9000, 9100)),
        }

        mp = multiprocessing.get_context("spawn")
        processes = []
        for i in range(nprocs):
            p = mp.Process(target=fn, args=(i, nranks, env, algorithm))
            p.start()
            processes.append(p)

        for p in processes:
            p.join(timeout=60)
            self.assertEqual(p.exitcode, 0)

    @skip_if_cuda_not_available()
    def test_gradient_allreduce(self):
        nprocs = torch.cuda.device_count()
        self.run_algorithm(
            nprocs, nprocs - 1, run_model_wrapper, algorithm="gradient_allreduce"
        )

    @skip_if_cuda_not_available()
    def test_decentralized(self):
        nprocs = torch.cuda.device_count()
        self.run_algorithm(
            nprocs, nprocs - 1, run_model_wrapper, algorithm="decentralized"
        )

    @skip_if_cuda_not_available()
    def test_low_prec_decentralized(self):
        nprocs = torch.cuda.device_count()
        self.run_algorithm(
            nprocs, nprocs - 1, run_model_wrapper, algorithm="low_prec_decentralized"
        )

    @skip_if_cuda_not_available()
    def test_async(self):
        nprocs = torch.cuda.device_count()
        self.run_algorithm(nprocs, nprocs - 1, run_model_wrapper, algorithm="async")

    @skip_if_cuda_not_available()
    def test_bytegrad(self):
        nprocs = torch.cuda.device_count()
        self.run_algorithm(nprocs, nprocs - 1, run_model_wrapper, algorithm="bytegrad")

    @skip_if_cuda_not_available()
    def test_qadam(self):
        nprocs = torch.cuda.device_count()
        self.run_algorithm(nprocs, nprocs - 1, run_model_wrapper, algorithm="qadam")

    @skip_if_cuda_not_available()
    def test_model_switch(self):
        nprocs = torch.cuda.device_count()
        algorithms = [
            "bytegrad",
            "decentralized",
            "gradient_allreduce",
            "qadam",
            "async",
            "low_prec_decentralized",
        ]
        self.run_algorithm(nprocs, nprocs - 1, run_model_switch_wrapper, algorithms)


if __name__ == "__main__":
    unittest.main()



import unittest
import torch
import torch.nn as nn
import torch.nn.functional as F
from bagua.torch_api.algorithms.q_adam import QAdamOptimizer
from tests import skip_if_cuda_available


class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.fc1 = nn.Linear(2, 10, bias=False)
        self.fc2 = nn.Linear(10, 50, bias=True)
        self.fc3 = nn.Linear(50, 4, bias=False)
        self.relu = nn.ReLU()

    def forward(self, x):
        x = self.relu(self.fc1(x))
        x = self.relu(self.fc2(x))
        x = self.fc3(x)
        return F.softmax(x, dim=1)


def run_step(opt_cls, opt_flags, seed):
    torch.manual_seed(seed)
    model = Net()
    optimizer = opt_cls(model.parameters(), **opt_flags)

    for step in range(1000):
        data = torch.randn(4, 2)
        target = torch.randn(4, 4)

        optimizer.zero_grad()
        output = model(data)
        loss = nn.MSELoss()(output, target)

        loss.backward()
        optimizer.step()

    return loss


class TestQAdam(unittest.TestCase):
    @skip_if_cuda_available()
    def test_qadam_optimizer(self):
        loss1 = run_step(torch.optim.Adam, {"lr": 0.001, "weight_decay": 0.1}, seed=13)
        loss2 = run_step(
            QAdamOptimizer,
            {"lr": 0.001, "weight_decay": 0.1, "warmup_steps": 2000},
            seed=13,
        )
        self.assertEqual(loss1.item(), loss2.item())



import bagua.torch_api as bagua
import torch
import torch.optim as optim
import unittest
import os
from tests.internal.common_utils import find_free_port
from tests import skip_if_cuda_available, skip_if_cuda_not_available

import logging

logging.getLogger().setLevel(logging.INFO)


def construct_model_and_optimizer(opt, flag_param, device):
    weight = torch.tensor(
        [[-0.2109, -0.4976], [-0.1413, -0.3420], [-0.2524, 0.6976]],
        requires_grad=True,
    )
    bias = torch.tensor(
        [-0.1085, -0.2979, 0.6892],
        requires_grad=True,
    )
    weight2 = torch.tensor(
        [[-0.0508, -0.3941, -0.2843]],
        requires_grad=True,
    )
    bias2 = torch.tensor([-0.0711], requires_grad=True)

    model = torch.nn.Sequential(
        torch.nn.Linear(2, 3),
        torch.nn.Sigmoid(),
        torch.nn.Linear(3, 1),
        torch.nn.Sigmoid(),
    )

    pretrained_dict = model.state_dict()
    pretrained_dict["0.weight"] = weight
    pretrained_dict["0.bias"] = bias
    pretrained_dict["2.weight"] = weight2
    pretrained_dict["2.bias"] = bias2
    model.load_state_dict(pretrained_dict)

    model = model.to(device)
    no_decay = ["bias"]
    optimizer_grouped_parameters = [
        {
            "params": [
                p
                for n, p in model.named_parameters()
                if not any(nd in n for nd in no_decay)
            ],
            "weight_decay": 0.1,
        },
        {
            "params": [
                p
                for n, p in model.named_parameters()
                if any(nd in n for nd in no_decay)
            ],
            "weight_decay": 0.0,
        },
    ]

    optimizer = opt(optimizer_grouped_parameters, **flag_param)

    return model, optimizer


def train_model(model, optimizer, device, num_epochs):
    input = torch.tensor([0.1, 0.2, 0.3, 0.4, 0.5, 0.6], device=device).reshape(3, 2)

    for epoch in range(num_epochs):
        optimizer.zero_grad()
        output = model(input)
        loss = output.sum()
        loss.backward()

        optimizer.step()
        # logging.debug(f"#train model#{epoch} params: {optimizer.param_groups}")
        # logging.debug(f"#train model#{epoch} state: {optimizer.state}")


def train_model_fused(model, optimizer, device, num_epochs):
    input = torch.tensor([0.1, 0.2, 0.3, 0.4, 0.5, 0.6], device=device).reshape(3, 2)

    for epoch in range(num_epochs):
        optimizer.zero_grad()
        output = model(input)
        loss = output.sum()
        loss.backward()

        if epoch % 2 == 0:
            optimizer.fuse_step()
        else:
            optimizer.step()
        # logging.debug(
        #    f"#train model fused#{epoch} params: {optimizer._bagua_fused_optimizer.param_groups}"
        # )
        # logging.debug(f"#train model fused#{epoch} state: {optimizer.state}")


def bagua_init(model, optimizer, algorithm, do_flatten):
    # wrap model
    if algorithm == "gradient_allreduce":
        from bagua.torch_api.algorithms import gradient_allreduce

        bagua_algorithm = gradient_allreduce.GradientAllReduceAlgorithm()
    elif algorithm == "bytegrad":
        from bagua.torch_api.algorithms import bytegrad

        bagua_algorithm = bytegrad.ByteGradAlgorithm()
    elif algorithm == "decentralized":
        from bagua.torch_api.algorithms import decentralized

        bagua_algorithm = decentralized.DecentralizedAlgorithm(hierarchical=False)
    elif algorithm == "async":
        from bagua.torch_api.algorithms import async_model_average

        bagua_algorithm = async_model_average.AsyncModelAverageAlgorithm(
            sync_interval_ms=10,
        )
    elif algorithm == "low_prec_decentralized":
        from bagua.torch_api.algorithms import decentralized

        bagua_algorithm = decentralized.LowPrecisionDecentralizedAlgorithm(
            hierarchical=False
        )
    elif algorithm == "qadam":
        from bagua.torch_api.algorithms.q_adam import QAdamAlgorithm, QAdamOptimizer

        optimizer = QAdamOptimizer(optimizer.param_groups, warmup_steps=1)
        bagua_algorithm = QAdamAlgorithm(optimizer, hierarchical=False)
    else:
        raise ValueError("unsupported algorithm")

    model = model.with_bagua([optimizer], bagua_algorithm, do_flatten=do_flatten)

    return model, optimizer


def setup_bagua_env():
    # init env
    os.environ["WORLD_SIZE"] = "1"
    os.environ["LOCAL_WORLD_SIZE"] = "1"
    os.environ["MASTER_ADDR"] = "127.0.0.1"
    os.environ["MASTER_PORT"] = str(find_free_port(8000, 8100))
    os.environ["BAGUA_SERVICE_PORT"] = str(find_free_port(9000, 9100))

    os.environ["RANK"] = "0"
    os.environ["LOCAL_RANK"] = "0"

    # init bagua distributed process group
    torch.cuda.set_device(0)
    # TODO: remove this after process group destroy supported
    if not bagua.communication.is_initialized():
        bagua.init_process_group()


def run(opt, flag_param, device, num_epochs):
    model, optimizer = construct_model_and_optimizer(opt, flag_param, device)

    train_model(model, optimizer, device, num_epochs=num_epochs)
    return model.parameters()


def run_fused(opt, flag_param, device, num_epochs):
    model, optimizer = construct_model_and_optimizer(opt, flag_param, device)
    optimizer = bagua.contrib.fuse_optimizer(optimizer, do_flatten=True)

    train_model_fused(model, optimizer, device, num_epochs=num_epochs)
    return model.parameters(), optimizer._bagua_fused_count


def run_with_bagua(opt, flag_param, device, num_epochs, algorithm):
    model, optimizer = construct_model_and_optimizer(opt, flag_param, device)

    model, optimizer = bagua_init(model, optimizer, algorithm, do_flatten=True)

    train_model(model, optimizer, device, num_epochs=num_epochs)

    if algorithm == "async":
        model.bagua_algorithm.abort(model)
    return model.parameters()


def run_fused_with_bagua(
    opt, flag_param, device, num_epochs, algorithm, optimizer_flatten, bagua_flatten
):
    model, optimizer = construct_model_and_optimizer(opt, flag_param, device)

    # First fuse optimizer, then wrap module
    optimizer = bagua.contrib.fuse_optimizer(optimizer, do_flatten=optimizer_flatten)
    model, optimizer = bagua_init(model, optimizer, algorithm, bagua_flatten)

    train_model_fused(model, optimizer, device, num_epochs=num_epochs)
    # torch.cuda.current_stream().synchronize()
    if algorithm == "async":
        model.bagua_algorithm.abort(model)
    # torch.cuda.synchronize()
    return model.parameters(), optimizer._bagua_fused_count


def run_fused_with_bagua_v2(
    opt, flag_param, device, num_epochs, algorithm, optimizer_flatten, bagua_flatten
):
    model, optimizer = construct_model_and_optimizer(opt, flag_param, device)

    # First wrap module, then fuse optimizer
    model, optimizer = bagua_init(model, optimizer, algorithm, bagua_flatten)
    optimizer = bagua.contrib.fuse_optimizer(optimizer, do_flatten=optimizer_flatten)

    train_model_fused(model, optimizer, device, num_epochs=num_epochs)

    if algorithm == "async":
        model.bagua_algorithm.abort(model)
    return model.parameters(), optimizer._bagua_fused_count


class TestFusedOptimizer(unittest.TestCase):
    def run_qadam(
        self, device, num_epochs, fused_count, optimizer_flatten, bagua_flatten
    ):
        res1 = run_with_bagua(
            optim.SGD,
            dict(lr=0.01),
            device=device,
            num_epochs=num_epochs,
            algorithm="qadam",
        )
        res2, cnt2 = run_fused_with_bagua_v2(
            optim.SGD,
            dict(lr=0.01),
            device=device,
            num_epochs=num_epochs,
            algorithm="qadam",
            optimizer_flatten=optimizer_flatten,
            bagua_flatten=bagua_flatten,
        )

        for p1, p2 in zip(res1, res2):
            self.assertTrue(torch.equal(p1, p2))
        self.assertTrue(cnt2 == fused_count)

    def run_all_optimizers_once(self, fn1, fn2, device, num_epochs, fused_count):
        optimizer_list = [
            optim.SGD,
            optim.SGD,
            optim.Adam,
            optim.Adam,
            optim.Adam,
            optim.Adam,
            optim.AdamW,
            optim.AdamW,
            optim.AdamW,
            optim.AdamW,
            optim.RMSprop,
            optim.RMSprop,
            optim.RMSprop,
            optim.RMSprop,
            optim.Rprop,
            optim.ASGD,
            optim.ASGD,
            optim.Adamax,
            optim.Adamax,
            optim.Adadelta,
            optim.Adadelta,
        ]

        flag_params = [
            dict(lr=0.2, momentum=1, dampening=0, weight_decay=1, nesterov=True),  # SGD
            dict(
                lr=0.2, momentum=1, dampening=0.5, weight_decay=1, nesterov=False
            ),  # SGD
            dict(weight_decay=1.0, amsgrad=True),  # Adam
            dict(weight_decay=1.0, amsgrad=False),  # Adam
            dict(weight_decay=0.0, amsgrad=True),  # Adam
            dict(weight_decay=0.0, amsgrad=False),  # Adam
            dict(weight_decay=1.0, amsgrad=True),  # AdamW
            dict(weight_decay=1.0, amsgrad=False),  # AdamW
            dict(weight_decay=0.0, amsgrad=True),  # AdamW
            dict(weight_decay=0.0, amsgrad=False),  # AdamW
            dict(weight_decay=1, momentum=1, centered=True),  # RMSprop
            dict(weight_decay=1, momentum=0, centered=True),  # RMSprop
            dict(weight_decay=1, momentum=1, centered=False),  # RMSprop
            dict(weight_decay=0, momentum=1, centered=False),  # RMSprop
            dict(lr=1e-2, etas=(0.5, 1.2), step_sizes=(1e-6, 50)),  # Rprop
            dict(weight_decay=0),  # ASGD
            dict(weight_decay=1),  # ASGD
            dict(weight_decay=0),  # Adamax
            dict(weight_decay=1),  # Adamax
            dict(weight_decay=0),  # Adadelta
            dict(weight_decay=1),  # Adadelta
        ]

        count = 0
        for opt, flag_param in zip(optimizer_list, flag_params):
            res1 = fn1(opt, flag_param, device=device, num_epochs=num_epochs)
            res2, cnt2 = fn2(opt, flag_param, device=device, num_epochs=num_epochs)

            for p1, p2 in zip(res1, res2):
                self.assertTrue(torch.equal(p1, p2))
            self.assertTrue(cnt2 == fused_count)

            count += 1
            if count % 5 == 0:
                logging.info(f"Tests Passed [{count}/{len(optimizer_list)}]")

    def run_fused_with_bagua_wrapper(self, fn1, fn2, num_epochs, fused_count):
        self.run_all_optimizers_once(fn1, fn2, "cuda:0", num_epochs, fused_count)

    @skip_if_cuda_available()
    def test_fused_optimizer(self):
        self.run_all_optimizers_once(
            fn1=run, fn2=run_fused, device="cpu", num_epochs=101, fused_count=102
        )

    @skip_if_cuda_not_available()
    def test_gradient_allreduce(self):
        setup_bagua_env()
        # check: optimizer param groups is flattened, should fuse
        self.run_fused_with_bagua_wrapper(
            fn1=run,
            fn2=lambda p1, p2, device, num_epochs: run_fused_with_bagua(
                p1, p2, device, num_epochs, "gradient_allreduce", True, False
            ),
            num_epochs=101,
            fused_count=102,
        )
        # check: both are falttened, should not fuse
        self.run_fused_with_bagua_wrapper(
            fn1=run,
            fn2=lambda p1, p2, device, num_epochs: run_fused_with_bagua(
                p1, p2, device, num_epochs, "gradient_allreduce", True, True
            ),
            num_epochs=101,
            fused_count=0,
        )
        # check: bagua module is falttened, should not fuse
        self.run_fused_with_bagua_wrapper(
            fn1=run,
            fn2=lambda p1, p2, device, num_epochs: run_fused_with_bagua(
                p1, p2, device, num_epochs, "gradient_allreduce", False, True
            ),
            num_epochs=101,
            fused_count=0,
        )

    @skip_if_cuda_not_available()
    def test_bytegrad(self):
        setup_bagua_env()
        # check: optimizer param groups is flattened, should fuse
        self.run_fused_with_bagua_wrapper(
            fn1=lambda p1, p2, device, num_epochs: run_with_bagua(
                p1, p2, device, num_epochs, "bytegrad"
            ),
            fn2=lambda p1, p2, device, num_epochs: run_fused_with_bagua(
                p1, p2, device, num_epochs, "bytegrad", True, False
            ),
            num_epochs=101,
            fused_count=102,
        )

    @skip_if_cuda_not_available()
    def test_decentralized(self):
        setup_bagua_env()
        # check: optimizer param groups is flattened, should fuse
        self.run_fused_with_bagua_wrapper(
            fn1=run,
            fn2=lambda p1, p2, device, num_epochs: run_fused_with_bagua(
                p1, p2, device, num_epochs, "decentralized", True, False
            ),
            num_epochs=101,
            fused_count=102,
        )
        self.run_fused_with_bagua_wrapper(
            fn1=run,
            fn2=lambda p1, p2, device, num_epochs: run_fused_with_bagua(
                p1, p2, device, num_epochs, "decentralized", True, True
            ),
            num_epochs=101,
            fused_count=0,
        )
        self.run_fused_with_bagua_wrapper(
            fn1=run,
            fn2=lambda p1, p2, device, num_epochs: run_fused_with_bagua(
                p1, p2, device, num_epochs, "decentralized", False, True
            ),
            num_epochs=101,
            fused_count=0,
        )

    @skip_if_cuda_not_available()
    def test_async(self):
        return
        setup_bagua_env()
        self.run_fused_with_bagua_wrapper(
            fn1=run,
            fn2=lambda p1, p2, device, num_epochs: run_fused_with_bagua(
                p1, p2, device, num_epochs, "async", True, False
            ),
            num_epochs=101,
            fused_count=102,
        )
        self.run_fused_with_bagua_wrapper(
            fn1=run,
            fn2=lambda p1, p2, device, num_epochs: run_fused_with_bagua(
                p1, p2, device, num_epochs, "async", False, True
            ),
            num_epochs=101,
            fused_count=0,
        )

    @skip_if_cuda_not_available()
    def test_low_prec_decentralized(self):
        return
        setup_bagua_env()
        self.run_fused_with_bagua_wrapper(
            fn1=lambda p1, p2, device, num_epochs: run_with_bagua(
                p1, p2, device, num_epochs, "low_prec_decentralized"
            ),
            fn2=lambda p1, p2, device, num_epochs: run_fused_with_bagua(
                p1, p2, device, num_epochs, "low_prec_decentralized", True, False
            ),
            num_epochs=101,
            fused_count=102,
        )

    @skip_if_cuda_not_available()
    def test_qadam(self):
        setup_bagua_env()
        self.run_qadam(
            device="cuda:0",
            num_epochs=101,
            fused_count=102,
            optimizer_flatten=True,
            bagua_flatten=False,
        )

    @skip_if_cuda_available()
    def test_calculate_mutual_groups(self):
        from bagua.torch_api.contrib.fuse.optimizer import calculate_mutual_groups

        tensor = torch.rand(100)

        tensor_pieces = []
        for i in range(10):
            tensor_pieces.append(tensor[i * 10 : (i + 1) * 10])

        g1 = [
            tensor_pieces[3],
            tensor_pieces[1],
            tensor_pieces[2],
            tensor_pieces[0],
            tensor_pieces[8],
            tensor_pieces[9],
            torch.rand(10),
        ]
        g2 = [torch.rand(10) for _ in range(len(g1))]
        g3 = [
            tensor_pieces[3],
            tensor_pieces[1],
            tensor_pieces[2],
            tensor_pieces[0],
            torch.rand(10),
            torch.rand(10),
            torch.rand(10),
        ]
        g4 = [
            torch.rand(10),
            tensor_pieces[1],
            tensor_pieces[2],
            tensor_pieces[0],
            tensor_pieces[8],
            tensor_pieces[9],
            torch.rand(10),
        ]
        g5 = [
            tensor_pieces[3],
            tensor_pieces[1],
            tensor_pieces[2],
            tensor_pieces[0],
            tensor_pieces[8],
            tensor_pieces[9],
            torch.rand(10),
        ]

        ret = calculate_mutual_groups([g1, g2])
        self.assertTrue(ret == [])

        ret = calculate_mutual_groups([g1, g3])
        self.assertTrue(ret == [[3, 1, 2, 0]])

        ret = calculate_mutual_groups([g1, g4])
        self.assertTrue(ret == [[4, 5]])

        ret = calculate_mutual_groups([g1, g5])
        self.assertTrue(ret == [[3, 1, 2, 0], [4, 5]])


if __name__ == "__main__":
    unittest.main()



import os
import unittest
import multiprocessing
import itertools
import inspect
from multiprocessing import Manager
import logging
import bagua.torch_api as bagua
from tests.internal.common_utils import find_free_port
from tests import skip_if_cuda_not_available
import torch
import torch.nn as nn
import torch.nn.functional as F
from bagua.torch_api.data_parallel import DistributedDataParallel as DDP


class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.fc1 = nn.Linear(2, 10, bias=False)
        self.fc2 = nn.Linear(10, 50, bias=True)
        self.fc3 = nn.Linear(50, 4, bias=False)
        self.relu = nn.ReLU()

    def forward(self, x):
        x = self.relu(self.fc1(x))
        x = self.relu(self.fc2(x))
        x = self.fc3(x)
        return F.softmax(x, dim=1)


def _init_bagua_env(rank, env):
    # Set deterministic
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True
    torch.manual_seed(rank)
    # Initialize subprocess env
    os.environ["WORLD_SIZE"] = env["WORLD_SIZE"]
    os.environ["LOCAL_WORLD_SIZE"] = env["LOCAL_WORLD_SIZE"]
    os.environ["MASTER_ADDR"] = env["MASTER_ADDR"]
    os.environ["MASTER_PORT"] = env["MASTER_PORT"]
    os.environ["BAGUA_SERVICE_PORT"] = env["BAGUA_SERVICE_PORT"]

    os.environ["RANK"] = str(rank)
    os.environ["LOCAL_RANK"] = str(rank)

    # Init bagua distributed process group
    torch.cuda.set_device(rank)
    bagua.init_process_group()


def create_model_and_optimizer(opt_class, opt_param):
    # C_in, C_out = 3, 10
    model = Net().cuda()
    hyper_param = {
        k: v
        for k, v in opt_param.items()
        if k in inspect.getargspec(opt_class.__init__).args
    }
    optimizer = opt_class(model.parameters(), **hyper_param)
    return model, optimizer


def get_optimizer_param_values(optimizer):
    results = []
    state_dict = optimizer.state_dict()
    for group in state_dict["param_groups"]:
        for param_id in group["params"]:
            if param_id not in state_dict["state"]:
                continue
            params = sorted(state_dict["state"][param_id].items())
            for k, v in params:
                results.append(
                    (k, v.clone().detach().cpu().numpy() if torch.is_tensor(v) else v)
                )
    return results


def run_bagua_broad(rank, nprocs, bagua_params, envs, opt_class, opt_hyper_param):
    _init_bagua_env(rank, envs)

    model, bagua_optimizer = create_model_and_optimizer(
        opt_class, opt_hyper_param
    )

    for epoch in range(5):
        logging.debug("Training epoch {}".format(epoch))
        for _ in range(10):
            data = torch.randn(4, 2).cuda()
            target = torch.randn(4, 4).cuda()

            bagua_optimizer.zero_grad()
            output = model(data)
            loss = nn.MSELoss()(output, target)

            loss.backward()
            bagua_optimizer.step()

    from bagua.torch_api.algorithms import gradient_allreduce

    algorithm = gradient_allreduce.GradientAllReduceAlgorithm(hierarchical=True)
    ddp_model = DDP(model, optimizers=[bagua_optimizer], algorithm=algorithm)

    model_params = [
        (k, v.clone().detach().cpu().numpy())
        for k, v in sorted(ddp_model.state_dict().items())
    ]
    optimizer_params = get_optimizer_param_values(bagua_optimizer)

    # Put "model_params" in dimension 1, while "optimizer_params" in dimension 2.
    bagua_params[rank][0].extend(model_params)
    bagua_params[rank][1].extend(optimizer_params)


class Test_Broadcast_Module(unittest.TestCase):
    @skip_if_cuda_not_available()
    def test_broadcast_module(self):
        nprocs = torch.cuda.device_count()

        optimizers = [
            (optim_class.__name__, optim_class)
            for optim_class in [torch.optim.SGD, torch.optim.Adam, torch.optim.Rprop]
        ]

        optimizer_hyper_param = [
            dict(lr=0.2, momentum=0.9, weight_decay=0.1, centered=True),
            dict(lr=0.2),
        ]

        for (opt_name, opt_class), opt_hyper_param in itertools.product(
            optimizers, optimizer_hyper_param
        ):
            env = {
                "WORLD_SIZE": str(nprocs),
                "LOCAL_WORLD_SIZE": str(nprocs),
                "MASTER_ADDR": "127.0.0.1",
                "MASTER_PORT": str(find_free_port(8000, 8100)),
                "BAGUA_SERVICE_PORT": str(find_free_port(9000, 9100)),
            }
            with Manager() as manager:
                # For each rank, set a two dimensional list. One is used to save model_params,
                # while the second save optimizer_params.
                bagua_params = manager.list(
                    [[manager.list() for _ in range(2)] for _ in range(nprocs)]
                )
                mp = multiprocessing.get_context("spawn")
                processes = []
                for i in range(nprocs):
                    p = mp.Process(
                        target=run_bagua_broad,
                        args=(
                            i,
                            nprocs,
                            bagua_params,
                            env,
                            opt_class,
                            opt_hyper_param,
                        ),
                    )
                    p.start()
                    processes.append(p)
                for p in processes:
                    p.join(timeout=60)
                    self.assertTrue(p.exitcode == 0)
                for rank in range(0, nprocs):
                    # Both "model_params" and "optimizer_params" are saved in (name, tensor/scalar) form,
                    # so we need to assert the two dimensional separately.
                    # This is compare the "model_params".
                    for i in range(len(bagua_params[0][0])):
                        # assert name
                        self.assertEqual(
                            bagua_params[0][0][i][0],
                            bagua_params[rank][0][i][0],
                        )
                        # assert tensor
                        self.assertTrue(
                            torch.equal(
                                torch.tensor(
                                    bagua_params[0][0][i][1], dtype=torch.float
                                ),
                                torch.tensor(
                                    bagua_params[rank][0][i][1], dtype=torch.float
                                ),
                            )
                        )

                    if len(bagua_params[0][1]) != 0:
                        for j in range(len(bagua_params[0][1])):
                            # assert name
                            self.assertEqual(
                                bagua_params[0][1][j][0],
                                bagua_params[rank][1][j][0],
                            )
                            # assert tensor/scalar
                            if (
                                bagua_params[0][1][j][1] is None
                            ):  # this is for "torch.optim.sgd.SGD" and dict(lr=0.2)
                                continue
                            else:
                                self.assertTrue(
                                    torch.equal(
                                        torch.tensor(
                                            bagua_params[0][1][j][1], dtype=torch.float
                                        ),
                                        torch.tensor(
                                            bagua_params[rank][1][j][1],
                                            dtype=torch.float,
                                        ),
                                    )
                                )


if __name__ == "__main__":
    unittest.main()
